"""
This module defines some classes that are generally useful for defining a type system for a new
domain. We inherit the type logic in ``nltk.sem.logic`` and add some functionality on top of it
here. There are two main improvements:
1) Firstly, we allow defining multiple basic types with their own names (see ``NamedBasicType``).
2) Secondly, we allow defining function types that have placeholders in them (see
``PlaceholderType``).
We also extend NLTK's ``LogicParser`` to define a ``DynamicTypeLogicParser`` that knows how to deal
with the two improvements above.
"""
from typing import Dict, List, Optional, Set, Tuple, Union
from collections import defaultdict
import itertools

from overrides import overrides
from nltk.sem.logic import Expression, ApplicationExpression, ConstantExpression, LogicParser, Variable
from nltk.sem.logic import Type, BasicType, ComplexType as NltkComplexType, ANY_TYPE

from allennlp.common.util import START_SYMBOL


class ComplexType(NltkComplexType):
    """
    In NLTK, a ``ComplexType`` is a function.  These functions are curried, so if you need multiple
    arguments for your function you nest ``ComplexTypes``.  That currying makes things difficult
    for us, and we mitigate the problems by adding ``return_type`` and ``argument_type`` functions
    to ``ComplexType``.
    """
    def return_type(self) -> Type:
        """
        Gives the final return type for this function.  If the function takes a single argument,
        this is just ``self.second``.  If the function takes multiple arguments and returns a basic
        type, this should be the final ``.second`` after following all complex types.  That is the
        implementation here in the base class.  If you have a higher-order function that returns a
        function itself, you need to override this method.
        """
        return_type = self.second
        while isinstance(return_type, ComplexType):
            return_type = return_type.second
        return return_type

    def argument_types(self) -> List[Type]:
        """
        Gives the types of all arguments to this function.  For functions returning a basic type,
        we grab all ``.first`` types until ``.second`` is no longer a ``ComplexType``.  That logic
        is implemented here in the base class.  If you have a higher-order function that returns a
        function itself, you need to override this method.
        """
        arguments = [self.first]
        remaining_type = self.second
        while isinstance(remaining_type, ComplexType):
            arguments.append(remaining_type.first)
            remaining_type = remaining_type.second
        return arguments

    def substitute_any_type(self, basic_types: Set[BasicType]) -> List[Type]:
        """
        Takes a set of ``BasicTypes`` and replaces any instances of ``ANY_TYPE`` inside this
        complex type with each of those basic types.
        """
        substitutions = []
        for first_type in substitute_any_type(self.first, basic_types):
            for second_type in substitute_any_type(self.second, basic_types):
                substitutions.append(self.__class__(first_type, second_type))
        return substitutions


class HigherOrderType(ComplexType):
    """
    A higher-order function is a ``ComplexType`` that returns functions.  We just override
    ``return_type`` and ``argument_types`` to make sure that these types are correct.

    Parameters
    ----------
    num_arguments : ``int``
        How many arguments this function takes before returning a function.  We'll go through this
        many levels of nested ``ComplexTypes`` before returning the final ``.second`` as our return
        type.
    first : ``Type``
        Passed to NLTK's ComplexType.
    second : ``Type``
        Passed to NLTK's ComplexType.
    """
    def __init__(self, num_arguments: int, first: Type, second: Type) -> None:
        super().__init__(first, second)
        self.num_arguments = num_arguments

    @overrides
    def return_type(self) -> Type:
        return_type = self.second
        for _ in range(self.num_arguments - 1):
            return_type = return_type.second
        return return_type

    @overrides
    def argument_types(self) -> List[Type]:
        arguments = [self.first]
        remaining_type = self.second
        for _ in range(self.num_arguments - 1):
            arguments.append(remaining_type.first)
            remaining_type = remaining_type.second
        return arguments


class NamedBasicType(BasicType):
    """
    A ``BasicType`` that also takes the name of the type as an argument to its constructor. Type
    resolution uses the output of ``__str__`` as well, so basic types with different
    representations do not resolve against each other.

    Parameters
    ----------
    string_rep : ``str``
        String representation of the type.
    """
    def __init__(self, string_rep) -> None:
        self._string_rep = string_rep

    def __str__(self):
        # TODO (pradeep): This limits the number of basic types we can have to 26. We may want to
        # change this in the future if we extend to domains where we have more than 26 basic types.
        if self._string_rep == START_SYMBOL:
            return START_SYMBOL
        else:
            return self._string_rep.lower()[0]

    def str(self):
        return self._string_rep


class MultiMatchNamedBasicType(NamedBasicType):
    """
    A ``NamedBasicType`` that matches with any type within a list of ``BasicTypes`` that it takes
    as an additional argument during instantiation. We just override the ``matches`` method in
    ``BasicType`` to match against any of the types given by the list.

    Parameters
    ----------
    string_rep : ``str``
        String representation of the type, passed to super class.
    types_to_match : ``List[BasicType]``
        List of types that this type should match with.
    """
    def __init__(self,
                 string_rep,
                 types_to_match: List[BasicType]) -> None:
        super().__init__(string_rep)
        self.types_to_match = set(types_to_match)

    @overrides
    def matches(self, other):
        return super().matches(other) or other in self.types_to_match


class PlaceholderType(ComplexType):
    """
    ``PlaceholderType`` is a ``ComplexType`` that involves placeholders, and thus its type
    resolution is context sensitive. This is an abstract class for all placeholder types like
    reverse, and, or, argmax, etc.

    Note that ANY_TYPE in NLTK's type system doesn't work like a wild card. Once the type of a
    variable gets resolved to a specific type, NLTK changes the type of that variable to that
    specific type. Hence, what NLTK calls "ANY_TYPE", is essentially a "yet-to-be-decided" type.
    This is a problem because we may want the same variable to bind to different types within a
    logical form, and using ANY_TYPE for this purpose will cause a resolution failure. For example
    the count function may apply to both rows and cells in the same logical form, and making count
    of type ``ComplexType(ANY_TYPE, DATE_NUM_TYPE)`` will cause a resolution error. This class lets
    you define ``ComplexType`` s with placeholders that are actually wild cards.

    The subclasses of this abstract class need to do three things
    1) Override the property ``_signature`` to define the type signature (this is just the
    signature's string representation and will not affect type inference or checking). You will see
    this signature in action sequences.
    2) Override ``resolve`` to resolve the type appropriately (see the docstring in ``resolve`` for
    more information).
    3) Override ``get_application_type`` which returns the return type when this type is applied as
    a function to an argument of a specified type.  For example, if you defined a reverse type by
    inheriting from this class, ``get_application_type`` gets an argument of type ``<a,b>``, it
    should return ``<b,a>`` .
    """
    _signature: str = None

    @overrides
    def resolve(self, other: Type) -> Optional[Type]:
        """
        This method is central to type inference and checking. When a variable's type is being
        checked, we compare what we know of its type against what is expected of its type by its
        context. The expectation is provided as ``other``. We make sure that there are no
        contradictions between this type and other, and return an updated type which may be more
        specific than the original type.

        For example, say this type is of the function variable F in F(cell), and we start out with
        ``<?, d>`` (that is, it takes any type and returns ``d`` ). Now we have already resolved
        cell to be of type ``e`` . Then ``resolve`` gets called with ``other = <e, ?>`` , because
        we know F is a function that took a constant of type ``e`` . When we resolve ``<e, ?>``
        against ``<?, d>`` , there will not be a contradiction, because any type can be
        successfully resolved against ``?`` . Finally we return ``<e, d>`` as the resolved type.

        As a counter example, if we are trying to resolve ``<?, d>`` against ``<?, e>`` , the
        resolution fails, and in that case, this method returns ``None`` .

        Note that a successful resolution does not imply equality of types because of one of them
        may be ANY_TYPE, and so in the subclasses of this type, we explicitly resolve in both
        directions.
        """
        raise NotImplementedError

    def get_application_type(self, argument_type: Type) -> Type:
        """
        This method returns the resulting type when this type is applied as a function to an argument of
        the given type.
        """
        raise NotImplementedError

    @overrides
    def substitute_any_type(self, basic_types: Set[BasicType]) -> List[Type]:
        """
        Placeholders mess with substitutions, so even though this method is implemented in the
        superclass, we override it here with a ``NotImplementedError`` to be sure that subclasses
        think about what the right thing to do here is, and do it correctly.
        """
        raise NotImplementedError

    @overrides
    def __eq__(self, other) -> bool:
        return self.__class__ == other.__class__

    @overrides
    def matches(self, other) -> bool:
        # self == ANY_TYPE = True iff self.first == ANY_TYPE and self.second == ANY_TYPE.
        return self == other or self == ANY_TYPE or other == ANY_TYPE

    @overrides
    def __str__(self):
        if self == ANY_TYPE:
            # If the type remains unresolved, we return ? instead of its signature.
            return str(ANY_TYPE)
        else:
            return self._signature

    @overrides
    def str(self):
        if self == ANY_TYPE:
            return ANY_TYPE.str()
        else:
            return self._signature

    __hash__ = ComplexType.__hash__


class UnaryOpType(PlaceholderType):
    """
    ``UnaryOpType`` is a kind of ``PlaceholderType`` that takes an argument of any type and returns
    an expression of the same type.  ``identity`` is an example of this kind of function.  The type
    signature of ``UnaryOpType`` is <#1, #1>.

    Parameters
    ----------
    allowed_substitutions : ``Set[BasicType]``, optional (default=None)
        If given, this sets restrictions on the types that can be substituted.  That is, say you
        have a unary operation that is only permitted for numbers and dates, you can pass those in
        here, and we will only consider those types when calling :func:`substitute_any_type`.  If
        this is ``None``, all basic types are allowed.
    signature : ``str``, optional (default='<#1,#1>')
        The signature of the operation is what will appear in action sequences that include this
        type.  The default value is suitable for functions that apply to any type.  If you have a
        restricted set of allowed substitutions, you likely want to change the type signature to
        reflect that.
    """
    def __init__(self,
                 type_: BasicType = ANY_TYPE,
                 allowed_substitutions: Set[BasicType] = None,
                 signature: str = '<#1,#1>') -> None:
        super().__init__(type_, type_)
        self._allowed_substitutions = allowed_substitutions
        self._signature = signature

    @overrides
    def resolve(self, other) -> Optional[Type]:
        """See ``PlaceholderType.resolve``"""
        if not isinstance(other, NltkComplexType):
            return None
        other_first = other.first.resolve(other.second)
        if not other_first:
            return None
        other_second = other.second.resolve(other_first)
        if not other_second:
            return None
        return UnaryOpType(other_first, self._allowed_substitutions, self._signature)

    @overrides
    def get_application_type(self, argument_type: Type) -> Type:
        return argument_type

    @overrides
    def substitute_any_type(self, basic_types: Set[BasicType]) -> List[Type]:
        if self.first != ANY_TYPE:
            return [self]
        allowed_basic_types = self._allowed_substitutions if self._allowed_substitutions else basic_types
        return [UnaryOpType(basic_type, self._allowed_substitutions, self._signature)
                for basic_type in allowed_basic_types]


class BinaryOpType(PlaceholderType):
    """
    ``BinaryOpType`` is a function that takes two arguments of the same type and returns an
    argument of that type.  ``+``, ``-``, ``and`` and ``or`` are examples of this kind of function.
    The type signature of ``BinaryOpType`` is ``<#1,<#1,#1>>``.

    Parameters
    ----------
    allowed_substitutions : ``Set[BasicType]``, optional (default=None)
        If given, this sets restrictions on the types that can be substituted.  That is, say you
        have a unary operation that is only permitted for numbers and dates, you can pass those in
        here, and we will only consider those types when calling :func:`substitute_any_type`.  If
        this is ``None``, all basic types are allowed.
    signature : ``str``, optional (default='<#1,<#1,#1>>')
        The signature of the operation is what will appear in action sequences that include this
        type.  The default value is suitable for functions that apply to any type.  If you have a
        restricted set of allowed substitutions, you likely want to change the type signature to
        reflect that.
    """
    def __init__(self,
                 type_: BasicType = ANY_TYPE,
                 allowed_substitutions: Set[BasicType] = None,
                 signature: str = '<#1,<#1,#1>>') -> None:
        super().__init__(type_, ComplexType(type_, type_))
        self._allowed_substitutions = allowed_substitutions
        self._signature = signature

    @overrides
    def resolve(self, other: Type) -> Optional[Type]:
        """See ``PlaceholderType.resolve``"""
        if not isinstance(other, NltkComplexType):
            return None
        if not isinstance(other.second, NltkComplexType):
            return None
        other_first = other.first.resolve(other.second.first)
        if other_first is None:
            return None
        other_first = other_first.resolve(other.second.second)
        if not other_first:
            return None
        other_second = other.second.resolve(ComplexType(other_first, other_first))
        if not other_second:
            return None
        return BinaryOpType(other_first, self._allowed_substitutions, self._signature)

    @overrides
    def get_application_type(self, argument_type: Type) -> Type:
        return ComplexType(argument_type, argument_type)

    @overrides
    def substitute_any_type(self, basic_types: Set[BasicType]) -> List[Type]:
        if self.first != ANY_TYPE:
            return [self]
        allowed_basic_types = self._allowed_substitutions if self._allowed_substitutions else basic_types
        return [BinaryOpType(basic_type, self._allowed_substitutions, self._signature)
                for basic_type in allowed_basic_types]


class TypedConstantExpression(ConstantExpression):
    # pylint: disable=abstract-method
    """
    NLTK assumes all constants are of type ``EntityType`` (e) by default. We define this new class
    where we can pass a default type to the constructor and use that in the ``_set_type`` method.
    """
    def __init__(self, variable, default_type: Type) -> None:
        super(TypedConstantExpression, self).__init__(variable)
        self._default_type = default_type

    @overrides
    def _set_type(self, other_type=ANY_TYPE, signature=None) -> None:
        if other_type == ANY_TYPE:
            super(TypedConstantExpression, self)._set_type(self._default_type, signature)
        else:
            super(TypedConstantExpression, self)._set_type(other_type, signature)


class DynamicTypeApplicationExpression(ApplicationExpression):
    """
    NLTK's ``ApplicationExpression`` (which represents function applications like P(x)) has two
    limitations, which we overcome by inheriting from ``ApplicationExpression`` and overriding two
    methods.

    Firstly, ``ApplicationExpression`` does not handle the case where P's type involves
    placeholders (R, V, !=, etc.), which are special cases because their return types depend on the
    type of their arguments (x). We override the property ``type`` to redefine the type of the
    application.

    Secondly, NLTK's variables only bind to entities, and thus the variable types are 'e' by
    default. We get around this issue by replacing x with X, whose initial type is ANY_TYPE, and
    later gets resolved based on the type signature of the function whose scope the variable
    appears in. This variable binding operation is implemented by overriding ``_set_type`` below.
    """
    def __init__(self, function: Expression, argument: Expression, variables_with_placeholders: Set[str]) -> None:
        super(DynamicTypeApplicationExpression, self).__init__(function, argument)
        self._variables_with_placeholders = variables_with_placeholders

    @property
    def type(self):
        # This gets called when the tree is being built by ``LogicParser.parse``. So, we do not
        # have access to the type signatures yet. Thus, we need to look at the name of the function
        # to return the type.
        if not str(self.function) in self._variables_with_placeholders:
            return super(DynamicTypeApplicationExpression, self).type
        if self.function.type == ANY_TYPE:
            return ANY_TYPE
        argument_type = self.argument.type
        return self.function.type.get_application_type(argument_type)

    def _set_type(self, other_type: Type = ANY_TYPE, signature=None) -> None:
        """
        We override this method to do just one thing on top of ``ApplicationExpression._set_type``.
        In lambda expressions of the form /x F(x), where the function is F and the argument is x,
        we can use the type of F to infer the type of x. That is, if F is of type <a, b>, we can
        resolve the type of x against a. We do this as the additional step after setting the type
        of F(x).

        So why does NLTK not already do this? NLTK assumes all variables (x) are of type entity
        (e).  So it does not have to resolve the type of x anymore. However, this would cause type
        inference failures in our case since x can bind to rows, numbers or cells, each of which
        has a different type. To deal with this issue, we made X of type ANY_TYPE. Also, LambdaDCS
        (and some other languages) contain a var function that indicate the usage of variables
        within lambda functions. We map var to V, and made it of type <#1, #1>. We cannot leave X
        as ANY_TYPE because that would propagate up the tree. We need to set its type when we have
        the information about F. Hence this method. Note that the language may or may not contain
        the var function. We deal with both cases below.
        """
        super(DynamicTypeApplicationExpression, self)._set_type(other_type, signature)
        # TODO(pradeep): Assuming the mapping of "var" function is "V". Do something better.
        if isinstance(self.argument, ApplicationExpression) and str(self.argument.function) == "V":
            # pylint: disable=protected-access
            self.argument.argument._set_type(self.function.type.first)
        if str(self.argument) == "X" and str(self.function) != "V":
            # pylint: disable=protected-access
            self.argument._set_type(self.function.type.first)


class DynamicTypeLogicParser(LogicParser):
    """
    ``DynamicTypeLogicParser`` is a ``LogicParser`` that can deal with ``NamedBasicType`` and
    ``PlaceholderType`` appropriately. Our extension here does two things differently.

    Firstly, we should handle constants of different types. We do this by passing a dict of format
    ``{name_prefix: type}`` to the constructor. For example, your domain has entities of types
    unicorns and elves, and you have an entity "Phil" of type unicorn, and "Bob" of type "elf". The
    names of the two entities should then be "unicorn:phil" and "elf:bob" respectively.

    Secondly, since we defined a new kind of ``ApplicationExpression`` above, the ``LogicParser``
    should be able to create this new kind of expression.
    """
    def __init__(self,
                 type_check: bool = True,
                 constant_type_prefixes: Dict[str, BasicType] = None,
                 type_signatures: Dict[str, Type] = None) -> None:
        super(DynamicTypeLogicParser, self).__init__(type_check)
        self._constant_type_prefixes = constant_type_prefixes or {}
        self._variables_with_placeholders = set([name for name, type_ in type_signatures.items()
                                                 if isinstance(type_, PlaceholderType)])

    @overrides
    def make_ApplicationExpression(self, function, argument):
        return DynamicTypeApplicationExpression(function, argument, self._variables_with_placeholders)

    @overrides
    def make_VariableExpression(self, name):
        if ":" in name:
            prefix = name.split(":")[0]
            if prefix in self._constant_type_prefixes:
                return TypedConstantExpression(Variable(name), self._constant_type_prefixes[prefix])
            else:
                raise RuntimeError(f"Unknown prefix: {prefix}. Did you forget to pass it to the constructor?")
        return super(DynamicTypeLogicParser, self).make_VariableExpression(name)

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented


class NameMapper:
    """
    The ``LogicParser`` we use has some naming conventions for functions (i.e. they should start
    with an upper case letter, and the remaining characters can only be digits). This means that we
    have to internally represent functions with unintuitive names. This class will automatically
    give unique names following the convention, and populate central mappings with these names. If
    for some reason you need to manually define the alias, you can do so by passing an alias to
    `map_name_with_signature`.

    Parameters
    ----------
    language_has_lambda : ``bool`` (optional, default=False)
        If your language has lambda functions, the word "lambda" needs to be in the name mapping,
        mapped to the alias "\". NLTK understands this symbol, and it doesn't need a type signature
        for it. Setting this flag to True adds the mapping to `name_mapping`.
    alias_prefix : ``str`` (optional, default="F")
        The one letter prefix used for all aliases. You do not need to specify it if you have only
        instance of this class for you language. If not, you can specify a different prefix for each
        name mapping you use for your language.
    """
    def __init__(self,
                 language_has_lambda: bool = False,
                 alias_prefix: str = "F") -> None:
        self.name_mapping: Dict[str, str] = {}
        if language_has_lambda:
            self.name_mapping["lambda"] = "\\"
        self.type_signatures: Dict[str, Type] = {}
        assert len(alias_prefix) == 1 and alias_prefix.isalpha(), (f"Invalid alias prefix: {alias_prefix}"
                                                                   "Needs to be a single upper case character.")
        self._alias_prefix = alias_prefix.upper()
        self._name_counter = 0

    def map_name_with_signature(self,
                                name: str,
                                signature: Type,
                                alias: str = None) -> None:
        if name in self.name_mapping:
            alias = self.name_mapping[name]
            old_signature = self.type_signatures[alias]
            if old_signature != signature:
                raise RuntimeError(f"{name} already added with signature {old_signature}. "
                                   f"Cannot add it again with {signature}!")
        else:
            alias = alias or f"{self._alias_prefix}{self._name_counter}"
            self._name_counter += 1
            self.name_mapping[name] = alias
            self.type_signatures[alias] = signature

    def get_alias(self, name: str) -> str:
        if name not in self.name_mapping:
            raise RuntimeError(f"Unmapped name: {name}")
        return self.name_mapping[name]

    def get_signature(self, name: str) -> Type:
        alias = self.get_alias(name)
        return self.type_signatures[alias]


def substitute_any_type(type_: Type, basic_types: Set[BasicType]) -> List[Type]:
    """
    Takes a type and a set of basic types, and substitutes all instances of ANY_TYPE with all
    possible basic types and returns a list with all possible combinations.  Note that this
    substitution is unconstrained.  That is, If you have a type with placeholders, <#1,#1> for
    example, this may substitute the placeholders with different basic types. In that case, you'd
    want to use ``_substitute_placeholder_type`` instead.
    """
    if type_ == ANY_TYPE:
        return list(basic_types)
    if isinstance(type_, BasicType):
        return [type_]
    # If we've made it this far, we have a ComplexType, and we can just call
    # `type_.substitute_any_type()`.
    return type_.substitute_any_type(basic_types)


def _make_production_string(source: Type, target: Union[List[Type], Type]) -> str:
    return f"{source} -> {target}"


def _get_complex_type_production(complex_type: ComplexType,
                                 multi_match_mapping: Dict[Type, List[Type]]) -> List[Tuple[Type, str]]:
    """
    Takes a complex type (without any placeholders), gets its return values, and returns productions
    (perhaps each with multiple arguments) that produce the return values.  This method also takes
    care of ``MultiMatchNamedBasicTypes``. If one of the arguments or the return types is a multi
    match type, it gets all the substitutions of those types from ``multi_match_mapping`` and forms
    a list with all possible combinations of substitutions. If the complex type passed to this method
    has no ``MultiMatchNamedBasicTypes``, the returned list will contain a single tuple.  For
    example, if the complex is type ``<a,<<b,c>,d>>``, and ``a`` is a multi match type that matches
    ``e`` and ``f``, this gives the following list of tuples: ``[('d', 'd -> [<a,<<b,c>,d>, e,
    <b,c>]), ('d', 'd -> [<a,<<b,c>,d>, f, <b,c>])]`` Note that we assume there will be no
    productions from the multi match type, and the list above does not contain ``('d', 'd ->
    [<a,<<b,c>,d>, a, <b,c>>]')``.
    """
    return_type = complex_type.return_type()
    if isinstance(return_type, MultiMatchNamedBasicType):
        return_types_matched = list(multi_match_mapping[return_type] if return_type in
                                    multi_match_mapping else return_type.types_to_match)
    else:
        return_types_matched = [return_type]
    arguments = complex_type.argument_types()
    argument_types_matched = []
    for argument_type in arguments:
        if isinstance(argument_type, MultiMatchNamedBasicType):
            matched_types = list(multi_match_mapping[argument_type] if argument_type in
                                 multi_match_mapping else argument_type.types_to_match)
            argument_types_matched.append(matched_types)
        else:
            argument_types_matched.append([argument_type])
    complex_type_productions: List[Tuple[Type, str]] = []
    for matched_return_type in return_types_matched:
        for matched_arguments in itertools.product(*argument_types_matched):
            complex_type_productions.append((matched_return_type,
                                             _make_production_string(return_type,
                                                                     [complex_type] + list(matched_arguments))))
    return complex_type_productions


def get_valid_actions(name_mapping: Dict[str, str],
                      type_signatures: Dict[str, Type],
                      basic_types: Set[Type],
                      multi_match_mapping: Dict[Type, List[Type]] = None,
                      valid_starting_types: Set[Type] = None,
                      num_nested_lambdas: int = 0) -> Dict[str, List[str]]:
    """
    Generates all the valid actions starting from each non-terminal. For terminals of a specific
    type, we simply add a production from the type to the terminal. For all terminal `functions`,
    we additionally add a rule that allows their return type to be generated from an application of
    the function.  For example, the function ``<e,<r,<d,r>>>``, which takes three arguments and
    returns an ``r`` would generate a the production rule ``r -> [<e,<r,<d,r>>>, e, r, d]``.

    For functions that do not contain ANY_TYPE or placeholder types, this is straight-forward.
    When there are ANY_TYPES or placeholders, we substitute the ANY_TYPE with all possible basic
    types, and then produce a similar rule.  For example, the identity function, with type
    ``<#1,#1>`` and basic types ``e`` and ``r``,  would produce the rules ``e -> [<#1,#1>, e]`` and
    ``r -> [<#1,#1>, r]``.

    We additionally add a valid action from the start symbol to all ``valid_starting_types``.

    Parameters
    ----------
    name_mapping : ``Dict[str, str]``
        The mapping of names that appear in your logical form languages to their aliases for NLTK.
        If you are getting all valid actions for a type declaration, this can be the
        ``COMMON_NAME_MAPPING``.
    type_signatures : ``Dict[str, Type]``
        The mapping from name aliases to their types. If you are getting all valid actions for a
        type declaration, this can be the ``COMMON_TYPE_SIGNATURE``.
    basic_types : ``Set[Type]``
        Set of all basic types in the type declaration.
    multi_match_mapping : ``Dict[Type, List[Type]]`` (optional)
        A mapping from `MultiMatchNamedBasicTypes` to the types they can match. This may be
        different from the type's ``types_to_match`` field based on the context. While building action
        sequences that lead to complex types with ``MultiMatchNamedBasicTypes``, if a type does not
        occur in this mapping, the default set of ``types_to_match`` for that type will be used.
    valid_starting_types : ``Set[Type]``, optional
        These are the valid starting types for your grammar; e.g., what types are we allowed to
        parse expressions into?  We will add a "START -> TYPE" rule for each of these types.  If
        this is ``None``, we default to using ``basic_types``.
    num_nested_lambdas : ``int`` (optional)
        Does the language used permit lambda expressions?  And if so, how many nested lambdas do we
        need to worry about?  We'll add rules like "<r,d> -> ['lambda x', d]" for all complex
        types, where the variable is determined by the number of nestings.  We currently only
        permit up to three levels of nesting, just for ease of implementation.
    """
    valid_actions: Dict[str, Set[str]] = defaultdict(set)

    valid_starting_types = valid_starting_types or basic_types
    for type_ in valid_starting_types:
        valid_actions[str(START_TYPE)].add(_make_production_string(START_TYPE, type_))

    complex_types = set()
    for name, alias in name_mapping.items():
        # Lambda functions and variables associated with them get produced in specific contexts. So
        # we do not add them to ``valid_actions`` here, and let ``GrammarState`` deal with it.
        # ``var`` is a special function that some languages (like LambdaDCS) use within lambda
        # functions to indicate the use of a variable (eg.: ``(lambda x (fb:row.row.year (var x)))``)
        # We do not have to produce this function outside the scope of lambda. Even within lambdas,
        # it is a lot easier to not do it, and let the action sequence to logical form transformation
        # logic add it to the output logical forms instead.
        if name in ["lambda", "var", "x", "y", "z"]:
            continue
        name_type = type_signatures[alias]
        # Type to terminal productions.
        for substituted_type in substitute_any_type(name_type, basic_types):
            valid_actions[str(substituted_type)].add(_make_production_string(substituted_type, name))
        # Keeping track of complex types.
        if isinstance(name_type, ComplexType) and name_type != ANY_TYPE:
            complex_types.add(name_type)

    for complex_type in complex_types:
        for substituted_type in substitute_any_type(complex_type, basic_types):
            for head, production in _get_complex_type_production(substituted_type,
                                                                 multi_match_mapping or {}):
                valid_actions[str(head)].add(production)

    # We can produce complex types with a lambda expression, though we'll leave out
    # placeholder types for now.
    for i in range(num_nested_lambdas):
        lambda_var = chr(ord('x') + i)
        # We'll only allow lambdas to be functions that take and return basic types as their
        # arguments, for now.  Also, we're doing this for all possible complex types where
        # the first and second types are basic types. So we may be overgenerating a bit.
        for first_type in basic_types:
            for second_type in basic_types:
                key = ComplexType(first_type, second_type)
                production_string = _make_production_string(key, ['lambda ' + lambda_var, second_type])
                valid_actions[str(key)].add(production_string)

    valid_action_strings = {key: sorted(value) for key, value in valid_actions.items()}
    return valid_action_strings


START_TYPE = NamedBasicType(START_SYMBOL)

# TODO(mattg): We're hard-coding three lambda variables here.  This isn't a great way to do
# this; it's just something that works for now, that we can fix later if / when it's needed.
# If you allow for more than three nested lambdas, or if you want to use different lambda
# variable names, you'll have to change this somehow.
LAMBDA_VARIABLES = set(['x', 'y', 'z'])

def is_nonterminal(production: str) -> bool:
    # TODO(pradeep): This is pretty specific to the assumptions made in converting types to
    # strings (e.g., that we're only using the first letter for types, lowercased).
    # TODO(pradeep): Also we simply check the surface forms here, and this works for
    # wikitables and nlvr. We should ideally let the individual type declarations define their own
    # variants of this method.
    if production in ['<=', '<']:
        # Some grammars (including the wikitables grammar) have "less than" and "less than or
        # equal to" functions that are terminals.  We don't want to treat those like our
        # "<t,d>" types.
        return False
    if production[0] == '<':
        return True
    if production.startswith('fb:'):
        return False
    if len(production) > 1 or production in LAMBDA_VARIABLES:
        return False
    return production[0].islower()



# pylint: disable=no-self-use,invalid-name,protected-access,too-many-public-methods
from allennlp.common.testing import AllenNlpTestCase
from allennlp.semparse.contexts import TableQuestionContext
from allennlp.data.tokenizers import WordTokenizer
from allennlp.data.tokenizers.word_splitter import SpacyWordSplitter


class TestTableQuestionContext(AllenNlpTestCase):
    def setUp(self):
        super().setUp()
        self.tokenizer = WordTokenizer(SpacyWordSplitter(pos_tags=True))

    def test_table_data(self):
        question = "what was the attendance when usl a league played?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/wikitables/sample_table.tagged'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        assert table_question_context.table_data == [{'date_column:year': '2001',
                                                      'number_column:division': '2',
                                                      'string_column:league': 'usl_a_league',
                                                      'string_column:regular_season': '4th_western',
                                                      'string_column:playoffs': 'quarterfinals',
                                                      'string_column:open_cup': 'did_not_qualify',
                                                      'number_column:avg_attendance': '7_169'},
                                                     {'date_column:year': '2005',
                                                      'number_column:division': '2',
                                                      'string_column:league': 'usl_first_division',
                                                      'string_column:regular_season': '5th',
                                                      'string_column:playoffs': 'quarterfinals',
                                                      'string_column:open_cup': '4th_round',
                                                      'number_column:avg_attendance': '6_028'}]

    def test_number_extraction(self):
        question = """how many players on the 191617 illinois fighting illini men's basketball team
                      had more than 100 points scored?"""
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-7.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        _, number_entities = table_question_context.get_entities_from_question()
        assert number_entities == [("191617", 5), ("100", 16)]

    def test_date_extraction(self):
        question = "how many laps did matt kenset complete on february 26, 2006."
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-8.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        _, number_entities = table_question_context.get_entities_from_question()
        assert number_entities == [("2", 8), ("26", 9), ("2006", 11)]

    def test_date_extraction_2(self):
        question = """how many different players scored for the san jose earthquakes during their
                      1979 home opener against the timbers?"""
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-6.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        _, number_entities = table_question_context.get_entities_from_question()
        assert number_entities == [("1979", 12)]

    def test_multiword_entity_extraction(self):
        question = "was the positioning better the year of the france venue or the year of the south korea venue?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-3.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        entities, _ = table_question_context.get_entities_from_question()
        assert entities == [("string:france", "string_column:venue"),
                            ("string:south_korea", "string_column:venue")]

    def test_rank_number_extraction(self):
        question = "what was the first tamil-language film in 1943?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-1.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        _, numbers = table_question_context.get_entities_from_question()
        assert numbers == [("1", 3), ('1943', 9)]

    def test_null_extraction(self):
        question = "on what date did the eagles score the least points?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-2.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        entities, numbers = table_question_context.get_entities_from_question()
        # "Eagles" does not appear in the table.
        assert entities == []
        assert numbers == []

    def test_numerical_column_type_extraction(self):
        question = """how many players on the 191617 illinois fighting illini men's basketball team
                      had more than 100 points scored?"""
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-7.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        predicted_types = table_question_context.column_types
        assert predicted_types["games_played"] == "number"
        assert predicted_types["field_goals"] == "number"
        assert predicted_types["free_throws"] == "number"
        assert predicted_types["points"] == "number"

    def test_date_column_type_extraction_1(self):
        question = "how many were elected?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-5.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        predicted_types = table_question_context.column_types
        assert predicted_types["first_elected"] == "date"

    def test_date_column_type_extraction_2(self):
        question = "how many were elected?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-9.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        predicted_types = table_question_context.column_types
        assert predicted_types["date_of_appointment"] == "date"
        assert predicted_types["date_of_election"] == "date"

    def test_string_column_types_extraction(self):
        question = "how many were elected?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-10.table'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        predicted_types = table_question_context.column_types
        assert predicted_types["birthplace"] == "string"
        assert predicted_types["advocate"] == "string"
        assert predicted_types["notability"] == "string"
        assert predicted_types["name"] == "string"

    def test_number_and_entity_extraction(self):
        question = "other than m1 how many notations have 1 in them?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f"{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-11.table"
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        string_entities, number_entities = table_question_context.get_entities_from_question()
        assert string_entities == [("string:m1", "string_column:notation")]
        assert number_entities == [("1", 2), ("1", 7)]

    def test_get_knowledge_graph(self):
        question = "other than m1 how many notations have 1 in them?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f"{self.FIXTURES_ROOT}/data/corenlp_processed_tables/TEST-11.table"
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        knowledge_graph = table_question_context.get_table_knowledge_graph()
        entities = knowledge_graph.entities
        # -1 is not in entities because there are no date columns in the table.
        assert sorted(entities) == ['1', 'number_column:position', 'string:m1',
                                    'string_column:mnemonic', 'string_column:notation',
                                    'string_column:short_name', 'string_column:swara']
        neighbors = knowledge_graph.neighbors
        # Each number extracted from the question will have all number and date columns as
        # neighbors. Each string entity extracted from the question will only have the corresponding
        # column as the neighbor.
        assert neighbors == {'1': ['number_column:position'],
                             'string_column:mnemonic': [],
                             'string_column:short_name': [],
                             'string_column:swara': [],
                             'number_column:position': ['1'],
                             'string:m1': ['string_column:notation'],
                             'string_column:notation': ['string:m1']}
        entity_text = knowledge_graph.entity_text
        assert entity_text == {'1': '1',
                               'string:m1': 'm1',
                               'string_column:notation': 'notation',
                               'string_column:mnemonic': 'mnemonic',
                               'string_column:short_name': 'short name',
                               'string_column:swara': 'swara',
                               'number_column:position': 'position'}


    def test_knowledge_graph_has_correct_neighbors(self):
        question = "when was the attendance greater than 5000?"
        question_tokens = self.tokenizer.tokenize(question)
        test_file = f'{self.FIXTURES_ROOT}/data/wikitables/sample_table.tagged'
        table_question_context = TableQuestionContext.read_from_file(test_file, question_tokens)
        knowledge_graph = table_question_context.get_table_knowledge_graph()
        neighbors = knowledge_graph.neighbors
        # '5000' is neighbors with number and date columns. '-1' is in entities because there is a
        # date column, which is its only neighbor.
        assert set(neighbors.keys()) == {'date_column:year', 'number_column:division',
                                         'string_column:league', 'string_column:regular_season',
                                         'string_column:playoffs', 'string_column:open_cup',
                                         'number_column:avg_attendance', '5000', '-1'}
        assert set(neighbors['date_column:year']) == {'5000', '-1'}
        assert neighbors['number_column:division'] == ['5000']
        assert neighbors['string_column:league'] == []
        assert neighbors['string_column:regular_season'] == []
        assert neighbors['string_column:playoffs'] == []
        assert neighbors['string_column:open_cup'] == []
        assert neighbors['number_column:avg_attendance'] == ['5000']
        assert set(neighbors['5000']) == {'date_column:year', 'number_column:division',
                                          'number_column:avg_attendance'}
        assert neighbors['-1'] == ['date_column:year']



"""
An implementation of the OpenAI Transformer Language Model.

Mostly just a slightly modified version of
https://github.com/huggingface/pytorch-openai-transformer-lm
so thanks to them!

Some of these modules duplicate code elsewhere in AllenNLP,
but the serialized weights depend on the exact parameter setup
here, so it's easiest to just reimplement them.
"""

# pylint: disable=invalid-name,arguments-differ
from typing import NamedTuple, List
import copy
import io
import json
import logging
import math
import pathlib
import re
import tarfile

import numpy as np
import torch
from torch.nn import Parameter

from allennlp.common.checks import ConfigurationError
from allennlp.common.file_utils import cached_path
from allennlp.common.from_params import FromParams

logger = logging.getLogger(__name__)

def gelu(x: torch.Tensor) -> torch.Tensor:
    return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))

def swish(x: torch.Tensor) -> torch.Tensor:
    return x * torch.sigmoid(x)

_ACTIVATION_FUNCTIONS = {
        'relu': torch.nn.ReLU,
        'swish': swish,
        'gelu': gelu
}

# pylint: disable=line-too-long
_PARAMETER_NAMES = ["model/we:0",
                    "model/h0/attn/c_attn/w:0", "model/h0/attn/c_attn/b:0", "model/h0/attn/c_proj/w:0",
                    "model/h0/attn/c_proj/b:0", "model/h0/ln_1/g:0", "model/h0/ln_1/b:0", "model/h0/mlp/c_fc/w:0",
                    "model/h0/mlp/c_fc/b:0", "model/h0/mlp/c_proj/w:0", "model/h0/mlp/c_proj/b:0", "model/h0/ln_2/g:0",
                    "model/h0/ln_2/b:0", "model/h1/attn/c_attn/w:0", "model/h1/attn/c_attn/b:0", "model/h1/attn/c_proj/w:0",
                    "model/h1/attn/c_proj/b:0", "model/h1/ln_1/g:0", "model/h1/ln_1/b:0", "model/h1/mlp/c_fc/w:0",
                    "model/h1/mlp/c_fc/b:0", "model/h1/mlp/c_proj/w:0", "model/h1/mlp/c_proj/b:0", "model/h1/ln_2/g:0",
                    "model/h1/ln_2/b:0", "model/h2/attn/c_attn/w:0", "model/h2/attn/c_attn/b:0", "model/h2/attn/c_proj/w:0",
                    "model/h2/attn/c_proj/b:0", "model/h2/ln_1/g:0", "model/h2/ln_1/b:0", "model/h2/mlp/c_fc/w:0",
                    "model/h2/mlp/c_fc/b:0", "model/h2/mlp/c_proj/w:0", "model/h2/mlp/c_proj/b:0", "model/h2/ln_2/g:0",
                    "model/h2/ln_2/b:0", "model/h3/attn/c_attn/w:0", "model/h3/attn/c_attn/b:0", "model/h3/attn/c_proj/w:0",
                    "model/h3/attn/c_proj/b:0", "model/h3/ln_1/g:0", "model/h3/ln_1/b:0", "model/h3/mlp/c_fc/w:0",
                    "model/h3/mlp/c_fc/b:0", "model/h3/mlp/c_proj/w:0", "model/h3/mlp/c_proj/b:0", "model/h3/ln_2/g:0",
                    "model/h3/ln_2/b:0", "model/h4/attn/c_attn/w:0", "model/h4/attn/c_attn/b:0", "model/h4/attn/c_proj/w:0",
                    "model/h4/attn/c_proj/b:0", "model/h4/ln_1/g:0", "model/h4/ln_1/b:0", "model/h4/mlp/c_fc/w:0",
                    "model/h4/mlp/c_fc/b:0", "model/h4/mlp/c_proj/w:0", "model/h4/mlp/c_proj/b:0", "model/h4/ln_2/g:0",
                    "model/h4/ln_2/b:0", "model/h5/attn/c_attn/w:0", "model/h5/attn/c_attn/b:0", "model/h5/attn/c_proj/w:0",
                    "model/h5/attn/c_proj/b:0", "model/h5/ln_1/g:0", "model/h5/ln_1/b:0", "model/h5/mlp/c_fc/w:0",
                    "model/h5/mlp/c_fc/b:0", "model/h5/mlp/c_proj/w:0", "model/h5/mlp/c_proj/b:0", "model/h5/ln_2/g:0",
                    "model/h5/ln_2/b:0", "model/h6/attn/c_attn/w:0", "model/h6/attn/c_attn/b:0", "model/h6/attn/c_proj/w:0",
                    "model/h6/attn/c_proj/b:0", "model/h6/ln_1/g:0", "model/h6/ln_1/b:0", "model/h6/mlp/c_fc/w:0",
                    "model/h6/mlp/c_fc/b:0", "model/h6/mlp/c_proj/w:0", "model/h6/mlp/c_proj/b:0", "model/h6/ln_2/g:0",
                    "model/h6/ln_2/b:0", "model/h7/attn/c_attn/w:0", "model/h7/attn/c_attn/b:0", "model/h7/attn/c_proj/w:0",
                    "model/h7/attn/c_proj/b:0", "model/h7/ln_1/g:0", "model/h7/ln_1/b:0", "model/h7/mlp/c_fc/w:0",
                    "model/h7/mlp/c_fc/b:0", "model/h7/mlp/c_proj/w:0", "model/h7/mlp/c_proj/b:0", "model/h7/ln_2/g:0",
                    "model/h7/ln_2/b:0", "model/h8/attn/c_attn/w:0", "model/h8/attn/c_attn/b:0", "model/h8/attn/c_proj/w:0",
                    "model/h8/attn/c_proj/b:0", "model/h8/ln_1/g:0", "model/h8/ln_1/b:0", "model/h8/mlp/c_fc/w:0",
                    "model/h8/mlp/c_fc/b:0", "model/h8/mlp/c_proj/w:0", "model/h8/mlp/c_proj/b:0", "model/h8/ln_2/g:0",
                    "model/h8/ln_2/b:0", "model/h9/attn/c_attn/w:0", "model/h9/attn/c_attn/b:0", "model/h9/attn/c_proj/w:0",
                    "model/h9/attn/c_proj/b:0", "model/h9/ln_1/g:0", "model/h9/ln_1/b:0", "model/h9/mlp/c_fc/w:0",
                    "model/h9/mlp/c_fc/b:0", "model/h9/mlp/c_proj/w:0", "model/h9/mlp/c_proj/b:0", "model/h9/ln_2/g:0",
                    "model/h9/ln_2/b:0", "model/h10/attn/c_attn/w:0", "model/h10/attn/c_attn/b:0", "model/h10/attn/c_proj/w:0",
                    "model/h10/attn/c_proj/b:0", "model/h10/ln_1/g:0", "model/h10/ln_1/b:0", "model/h10/mlp/c_fc/w:0",
                    "model/h10/mlp/c_fc/b:0", "model/h10/mlp/c_proj/w:0", "model/h10/mlp/c_proj/b:0", "model/h10/ln_2/g:0",
                    "model/h10/ln_2/b:0", "model/h11/attn/c_attn/w:0", "model/h11/attn/c_attn/b:0", "model/h11/attn/c_proj/w:0",
                    "model/h11/attn/c_proj/b:0", "model/h11/ln_1/g:0", "model/h11/ln_1/b:0", "model/h11/mlp/c_fc/w:0",
                    "model/h11/mlp/c_fc/b:0", "model/h11/mlp/c_proj/w:0", "model/h11/mlp/c_proj/b:0", "model/h11/ln_2/g:0",
                    "model/h11/ln_2/b:0", "model/clf/w:0", "model/clf/b:0"]
# pylint: enable=line-too-long


class TransformerConfig(NamedTuple):
    """
    The transformer has to pass a bunch of params to its submodules,
    this bundles them together to make things easier.
    """
    embedding_dim: int = 768
    num_heads: int = 12
    embedding_dropout_probability: float = 0.1
    attention_dropout_probability: float = 0.1
    residual_dropout_probability: float = 0.1
    activation_function: str = 'gelu'


class LayerNorm(torch.nn.Module):
    """Construct a layernorm module in the OpenAI style (epsilon inside the square root)."""

    def __init__(self, n_state, e=1e-5):
        super().__init__()
        self.g = torch.nn.Parameter(torch.ones(n_state))
        self.b = torch.nn.Parameter(torch.zeros(n_state))
        self.e = e

    def forward(self, x):
        u = x.mean(-1, keepdim=True)
        s = (x - u).pow(2).mean(-1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.e)
        return self.g * x + self.b


class Conv1D(torch.nn.Module):
    def __init__(self, nf: int, rf: int, nx: int) -> None:
        super().__init__()
        self.rf = rf
        self.nf = nf
        if rf == 1:
            w = torch.empty(nx, nf)
            torch.nn.init.normal_(w, std=0.02)
            self.w = Parameter(w)
            self.b = Parameter(torch.zeros(nf))
        else:
            raise NotImplementedError

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.rf == 1:
            size_out = x.size()[:-1] + (self.nf,)
            x = torch.addmm(self.b, x.view(-1, x.size(-1)), self.w)
            x = x.view(*size_out)
        else:
            raise NotImplementedError
        return x

class Attention(torch.nn.Module):
    def __init__(self,
                 nx: int,
                 n_ctx: int,
                 config: TransformerConfig,
                 scale: bool = False) -> None:
        super().__init__()
        n_state = nx  # in Attention: n_state=768 (nx=n_embd)
        # [switch nx => n_state from Block to Attention to keep identical to TF implem]
        assert n_state % config.num_heads == 0
        self.register_buffer('b', torch.tril(torch.ones(n_ctx, n_ctx)).view(1, 1, n_ctx, n_ctx))
        self.n_head = config.num_heads
        self.split_size = n_state
        self.scale = scale
        self.c_attn = Conv1D(n_state * 3, 1, nx)
        self.c_proj = Conv1D(n_state, 1, nx)
        self.attn_dropout = torch.nn.Dropout(config.attention_dropout_probability)
        self.resid_dropout = torch.nn.Dropout(config.residual_dropout_probability)

    def _attn(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
        w = torch.matmul(q, k)
        if self.scale:
            w = w / math.sqrt(v.size(-1))
        w = w * self.b + -1e9 * (1 - self.b)  # TF implem method: mask_attn_weights
        w = torch.nn.Softmax(dim=-1)(w)
        w = self.attn_dropout(w)
        return torch.matmul(w, v)

    def merge_heads(self, x: torch.Tensor):
        # pylint: disable=no-self-use
        x = x.permute(0, 2, 1, 3).contiguous()
        new_x_shape = x.size()[:-2] + (x.size(-2) * x.size(-1),)
        return x.view(*new_x_shape)  # in Tensorflow implem: fct merge_states

    def split_heads(self, x: torch.Tensor, k: bool = False):
        new_x_shape = x.size()[:-1] + (self.n_head, x.size(-1) // self.n_head)
        x = x.view(*new_x_shape)  # in Tensorflow implem: fct split_states
        if k:
            return x.permute(0, 2, 3, 1)
        else:
            return x.permute(0, 2, 1, 3)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.c_attn(x)
        query, key, value = x.split(self.split_size, dim=2)
        query = self.split_heads(query)
        key = self.split_heads(key, k=True)
        value = self.split_heads(value)
        a = self._attn(query, key, value)
        a = self.merge_heads(a)
        a = self.c_proj(a)
        a = self.resid_dropout(a)
        return a


class MLP(torch.nn.Module):
    def __init__(self, n_state: int, config: TransformerConfig) -> None:  # in MLP: n_state=3072 (4 * n_embd)
        super().__init__()
        self.c_fc = Conv1D(n_state, 1, config.embedding_dim)
        self.c_proj = Conv1D(config.embedding_dim, 1, n_state)
        self.act = _ACTIVATION_FUNCTIONS[config.activation_function]
        self.dropout = torch.nn.Dropout(config.residual_dropout_probability)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        h = self.act(self.c_fc(x))
        h2 = self.c_proj(h)
        return self.dropout(h2)


class Block(torch.nn.Module):
    def __init__(self,
                 n_ctx: int,
                 config: TransformerConfig,
                 scale: bool = False) -> None:
        super().__init__()
        nx = config.embedding_dim
        self.attn = Attention(nx, n_ctx, config, scale)
        self.ln_1 = LayerNorm(nx)
        self.mlp = MLP(4 * nx, config)
        self.ln_2 = LayerNorm(nx)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = self.attn(x)
        n = self.ln_1(x + a)
        m = self.mlp(n)
        h = self.ln_2(n + m)
        return h

class OpenaiTransformer(torch.nn.Module, FromParams):
    """
    Openai transformer, as per https://blog.openai.com/language-unsupervised/.
    Default parameters are the ones for their pretrained model.

    Parameters
    ----------
    vocab_size: ``int`` (optional, default: 40478)
        The size of the vocabulary (number of byte pair embeddings)
        excluding the n_special embeddings (if any), and the positional embeddings.
    n_ctx: ``int`` (optional, default: 512)
        The number of positional encodings to use for evaluation.
    embedding_dim: ``int`` (optional, default: 768)
        The dimension of the output embeddings.
    num_heads: ``int`` (optional, default: 12)
        How many "heads" the attention has.
    num_layers: ``int`` (optional, default: 12)
        How many layers of "blocks" the transformer has.
    embedding_dropout_probability: ``float`` (optional, default: 0.1)
        Dropout for the embedding.
    attention_dropout_probability: ``float`` (optional, default: 0.1)
        Dropout for attention.
    residual_dropout_probability: ``float`` (optional, default: 0.1)
        Dropout for residual
    activation_function: ``str`` (optional, default: ``'gelu'``)
        Activation function for the multi-layer perceptron.
    model_path: ``str`` (optional, default: ``None``)
        A tar.gz file containing serialized model weights. If supplied,
        the weights will be loaded from that file.
    requires_grad: ``bool`` (optional, default: ``False``)
        If true, the transformer will be fine-tuneable.
    n_special: ``int`` (optional, default: ``-1``)
        The number of special tokens added to the byte pair vocabulary
        (via ``OpenaiTransformerBytePairIndexer``).
    """
    def __init__(self,
                 vocab_size: int = 40478,
                 n_ctx: int = 512,
                 embedding_dim: int = 768,
                 num_heads: int = 12,
                 num_layers: int = 12,
                 embedding_dropout_probability: float = 0.1,
                 attention_dropout_probability: float = 0.1,
                 residual_dropout_probability: float = 0.1,
                 activation_function: str = 'gelu',
                 model_path: str = None,
                 requires_grad: bool = False,
                 n_special: int = -1) -> None:
        super().__init__()

        config = TransformerConfig(
                embedding_dim,
                num_heads,
                embedding_dropout_probability,
                attention_dropout_probability,
                residual_dropout_probability,
                activation_function,
        )

        # the embedding size is vocab_size + n_special embeddings + n_ctx
        embedding_size = vocab_size + max(n_special, 0) + n_ctx
        self.vocab_size = embedding_size
        self.n_ctx = n_ctx
        self.n_special = n_special

        self.num_output_layers = 1 + num_layers

        self.embed = torch.nn.Embedding(embedding_size, embedding_dim)
        self.drop = torch.nn.Dropout(embedding_dropout_probability)

        block = Block(n_ctx, config, scale=True)
        self.h = torch.nn.ModuleList([copy.deepcopy(block) for _ in range(num_layers)])
        self.decoder = torch.nn.Linear(embedding_dim, embedding_size, bias=False)
        self.decoder.weight = self.embed.weight  # Tied weights
        # To reproduce the noise_shape parameter of TF implementation

        torch.nn.init.normal_(self.embed.weight, std=0.02)

        for parameter in self.parameters():
            parameter.requires_grad = requires_grad

        if model_path:
            self.load_weights(model_path, n_special=n_special, n_ctx=n_ctx)

    def forward(self, x: torch.Tensor) -> List[torch.Tensor]:
        #x = x.view(-1, x.size(2), x.size(3))

        # x is (batch_size, sequence_length) tensor of byte-pair ids

        # e is (batch_size, sequence_length, 2, embedding_dim) tensor of embeddings
        e = self.embed(x)

        # h is (batch_size, sequence_length, embedding_dim)
        h = e.sum(dim=2)

        all_layers = [h]
        for block in self.h:
            h = block(h)
            all_layers.append(h)

        # result is list of (batch_size, sequence_length, embedding_dim)
        return all_layers

    def load_weights(self,
                     transformer_model_path: str,
                     n_ctx: int = -1,
                     n_special: int = -1,
                     n_transfer: int = 12,
                     n_embd: int = 768,
                     names: List[str] = _PARAMETER_NAMES) -> None:
        # pylint: disable=dangerous-default-value

        logger.info(f"loading weights from {transformer_model_path}")
        # if `file_path` is a URL, redirect to the cache
        transformer_model_path = cached_path(transformer_model_path)

        with tarfile.open(transformer_model_path) as tmp:
            num_params_files = len([member for member in tmp.getmembers() if member.name.endswith('.npy')])
            shapesfile = tmp.extractfile('model/params_shapes.json')
            if shapesfile:
                shapes = json.loads(shapesfile.read())
            else:
                raise ConfigurationError("unable to find model/params_shapes.json in the archive")

            # numpy can't read from a tarfile directly, so we need a workaround
            # https://github.com/numpy/numpy/issues/7989#issuecomment-341656702
            init_params: List[np.ndarray] = []
            for n in range(num_params_files):
                array_file = io.BytesIO()
                array_file.write(tmp.extractfile(f'model/params_{n}.npy').read())
                array_file.seek(0)
                # each np.load is a (11653478,) numpy array
                init_params.append(np.load(array_file))

        # init_params is a list of 10 arrays of size (11653578,)
        # shapes are [[512, 768], [40478, 768], [1, 768, 2304], [2304], ...  # 146 elts
        # products are [512 * 768, 40478 * 768, ...]
        # offsets is [512 * 768, 512 * 768 + 40478 * 768, ...]
        offsets = np.cumsum([np.prod(shape) for shape in shapes])

        # split into the 146 subarrays corresponding to shapes
        init_params = np.split(np.concatenate(init_params, 0), offsets)[:-1]

        # reshape
        init_params = [param.reshape(shape) for param, shape in zip(init_params, shapes)]

        # truncate if necessary
        if n_ctx > 0:
            # positional embeddings?
            # init_params[0] is (512, 768) = (max_chars, embedding_dim)
            init_params[0] = init_params[0][:n_ctx]

        # combine init_params[1] and init_params[0]
        if n_special > 0:
            # init_params[1] is (40478, 768)
            # special is (n_special, 768)
            # init_params[0] is (512, 768)
            # result is (40990 + n_special, 768)
            init_params[0] = np.concatenate(
                    [init_params[1],
                     (np.random.randn(n_special, n_embd) * 0.02).astype(np.float32),
                     init_params[0]],
                    0
            )
        else:
            # result is (40990, 768)
            init_params[0] = np.concatenate([init_params[1], init_params[0]], 0)
        del init_params[1]

        # number of dimensions to transfer, 12 per layer, plus one extra
        if n_transfer == -1:
            n_transfer = 0
        else:
            n_transfer = 1 + n_transfer * 12

        # squeeze?
        init_params = [arr.squeeze() for arr in init_params]

        # embedding.weight is (vocab_size, embedding_dim)
        # make sure init_params[0] has the same shape
        try:
            assert self.embed.weight.shape == init_params[0].shape
        except AssertionError as e:
            e.args += (self.embed.weight.shape, init_params[0].shape)
            raise

        # and then assign it
        self.embed.weight.data = torch.from_numpy(init_params[0])
        self.decoder.weight = self.embed.weight

        # for each (name, array) pair to transfer over
        for name, ip in zip(names[1:n_transfer], init_params[1:n_transfer]):
                                            # "model/h0/attn/c_attn/w:0"
            name = name[6:]                 # "h0/attn/c_attn/w:0"
            assert name[-2:] == ":0"
            name = name[:-2]                # "h0/attn/c_attn/w"
            name_parts = name.split('/')    # ['h0', 'attn', 'c_attn', 'w']

            pointer = self
            for m_name in name_parts:
                if re.fullmatch(r'[A-Za-z]+\d+', m_name):
                    l = re.split(r'(\d+)', m_name)   # ['h', '0', '']
                else:
                    l = [m_name]                     # ['attn']
                pointer = getattr(pointer, l[0])
                if len(l) >= 2:
                    num = int(l[1])
                    pointer = pointer[num]
            try:
                assert pointer.shape == ip.shape
            except AssertionError as e:
                e.args += (pointer.shape, ip.shape)
                raise

            pointer.data = torch.from_numpy(ip)  # pylint: disable=attribute-defined-outside-init

    def dump_weights(self,
                     output_dir: str,
                     num_pieces: int = 10) -> None:
        output_path = pathlib.Path(output_dir) / 'model'
        output_path.mkdir(exist_ok=True, parents=True)  # pylint: disable=no-member

        named_parameters = list(self.named_parameters())

        # embedding weights get special treatment
        _, array = named_parameters[0]
        num_bpe = self.vocab_size - self.n_ctx
        byte_pair_embeddings = array[:num_bpe]
        positional_embeddings = array[num_bpe:]

        arrays = [positional_embeddings.numpy().ravel(), byte_pair_embeddings.numpy().ravel()]
        shapes = [positional_embeddings.shape, byte_pair_embeddings.shape]
        names = ["model/we:0"]

        for param_name, tensor in named_parameters[1:]:
            param_name = f'h{param_name}'            # 'h0.attn.c_attn.w'
            parts = param_name.split(".")            # ['h0', 'attn', 'c_attn', 'w']
            name = "model/" + '/'.join(parts) + ':0' # 'model/h0/attn/c_attn/w:0'
            array = tensor.numpy().ravel()

            arrays.append(array)
            shapes.append(list(tensor.shape))
            names.append(name)

        # write out the arrays
        big_array = np.concatenate(arrays)
        total_size = len(big_array)
        batch_size = math.ceil(total_size / num_pieces)

        for i in range(num_pieces):
            filename = output_path / f"params_{i}.npy"
            start = i * batch_size
            end = start + batch_size
            subarray = big_array[start:end]

            np.save(filename, subarray)

        # write out the shapes
        with open(output_path / 'params_shapes.json', 'w') as shapes_file:
            json.dump(shapes, shapes_file)



from typing import Dict, List, Union
import logging
import json
from overrides import overrides
from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import LabelField, TextField, Field, ListField
from allennlp.data.instance import Instance
from allennlp.data.token_indexers import TokenIndexer, SingleIdTokenIndexer
from allennlp.data.tokenizers import Tokenizer, WordTokenizer
from allennlp.data.tokenizers.sentence_splitter import SpacySentenceSplitter

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


@DatasetReader.register("text_classification_json")
class TextClassificationJsonReader(DatasetReader):
    """
    Reads tokens and their labels from a labeled text classification dataset.
    Expects a "text" field and a "label" field in JSON format.

    The output of ``read`` is a list of ``Instance`` s with the fields:
        tokens: ``TextField`` and
        label: ``LabelField``

    Parameters
    ----------
    token_indexers : ``Dict[str, TokenIndexer]``, optional
        optional (default=``{"tokens": SingleIdTokenIndexer()}``)
        We use this to define the input representation for the text.
        See :class:`TokenIndexer`.
    tokenizer : ``Tokenizer``, optional (default = ``{"tokens": WordTokenizer()}``)
        Tokenizer to use to split the input text into words or other kinds of tokens.
    segment_sentences: ``bool``, optional (default = ``False``)
        If True, we will first segment the text into sentences using SpaCy and then tokenize words.
        Necessary for some models that require pre-segmentation of sentences, like the Hierarchical
        Attention Network (https://www.cs.cmu.edu/~hovy/papers/16HLT-hierarchical-attention-networks.pdf).
    max_sequence_length: ``int``, optional (default = ``None``)
        If specified, will truncate tokens to specified maximum length.
    skip_label_indexing: ``bool``, optional (default = ``False``)
        Whether or not to skip label indexing. You might want to skip label indexing if your
        labels are numbers, so the dataset reader doesn't re-number them starting from 0.
    lazy : ``bool``, optional, (default = ``False``)
        Whether or not instances can be read lazily.
    """
    def __init__(self,
                 token_indexers: Dict[str, TokenIndexer] = None,
                 tokenizer: Tokenizer = None,
                 segment_sentences: bool = False,
                 max_sequence_length: int = None,
                 skip_label_indexing: bool = False,
                 lazy: bool = False) -> None:
        super().__init__(lazy=lazy)
        self._tokenizer = tokenizer or WordTokenizer()
        self._segment_sentences = segment_sentences
        self._max_sequence_length = max_sequence_length
        self._skip_label_indexing = skip_label_indexing
        self._token_indexers = token_indexers or {'tokens': SingleIdTokenIndexer()}
        if self._segment_sentences:
            self._sentence_segmenter = SpacySentenceSplitter()

    @overrides
    def _read(self, file_path):
        with open(cached_path(file_path), "r") as data_file:
            for line in data_file.readlines():
                if not line:
                    continue
                items = json.loads(line)
                text = items["text"]
                label = items.get("label", None)
                if label is not None:
                    if self._skip_label_indexing:
                        try:
                            label = int(label)
                        except ValueError:
                            raise ValueError('Labels must be integers if skip_label_indexing is True.')
                    else:
                        label = str(label)
                instance = self.text_to_instance(text=text, label=label)
                if instance is not None:
                    yield instance

    def _truncate(self, tokens):
        """
        truncate a set of tokens using the provided sequence length
        """
        if len(tokens) > self._max_sequence_length:
            tokens = tokens[:self._max_sequence_length]
        return tokens

    @overrides
    def text_to_instance(self, text: str, label: Union[str, int] = None) -> Instance:  # type: ignore
        """
        Parameters
        ----------
        text : ``str``, required.
            The text to classify
        label : ``str``, optional, (default = None).
            The label for this text.

        Returns
        -------
        An ``Instance`` containing the following fields:
            tokens : ``TextField``
                The tokens in the sentence or phrase.
            label : ``LabelField``
                The label label of the sentence or phrase.
        """
        # pylint: disable=arguments-differ
        fields: Dict[str, Field] = {}
        if self._segment_sentences:
            sentences: List[Field] = []
            sentence_splits = self._sentence_segmenter.split_sentences(text)
            for sentence in sentence_splits:
                word_tokens = self._tokenizer.tokenize(sentence)
                if self._max_sequence_length is not None:
                    word_tokens = self._truncate(word_tokens)
                sentences.append(TextField(word_tokens, self._token_indexers))
            fields['tokens'] = ListField(sentences)
        else:
            tokens = self._tokenizer.tokenize(text)
            if self._max_sequence_length is not None:
                tokens = self._truncate(tokens)
            fields['tokens'] = TextField(tokens, self._token_indexers)
        if label is not None:
            fields['label'] = LabelField(label,
                                         skip_indexing=self._skip_label_indexing)
        return Instance(fields)



import torch

from allennlp.common.checks import ConfigurationError
from allennlp.modules.seq2vec_encoders.seq2vec_encoder import Seq2VecEncoder


class PytorchSeq2VecWrapper(Seq2VecEncoder):
    """
    Pytorch's RNNs have two outputs: the hidden state for every time step, and the hidden state at
    the last time step for every layer.  We just want the second one as a single output.  This
    wrapper pulls out that output, and adds a :func:`get_output_dim` method, which is useful if you
    want to, e.g., define a linear + softmax layer on top of this to get some distribution over a
    set of labels.  The linear layer needs to know its input dimension before it is called, and you
    can get that from ``get_output_dim``.

    Also, there are lots of ways you could imagine going from an RNN hidden state at every
    timestep to a single vector - you could take the last vector at all layers in the stack, do
    some kind of pooling, take the last vector of the top layer in a stack, or many other  options.
    We just take the final hidden state vector, or in the case of a bidirectional RNN cell, we
    concatenate the forward and backward final states together. TODO(mattg): allow for other ways
    of wrapping RNNs.

    In order to be wrapped with this wrapper, a class must have the following members:

        - ``self.input_size: int``
        - ``self.hidden_size: int``
        - ``def forward(inputs: PackedSequence, hidden_state: torch.tensor) ->
          Tuple[PackedSequence, torch.Tensor]``.
        - ``self.bidirectional: bool`` (optional)

    This is what pytorch's RNN's look like - just make sure your class looks like those, and it
    should work.

    Note that we *require* you to pass sequence lengths when you call this module, to avoid subtle
    bugs around masking.  If you already have a ``PackedSequence`` you can pass ``None`` as the
    second parameter.
    """
    def __init__(self, module: torch.nn.modules.RNNBase) -> None:
        # Seq2VecEncoders cannot be stateful.
        super(PytorchSeq2VecWrapper, self).__init__(stateful=False)
        self._module = module
        try:
            if not self._module.batch_first:
                raise ConfigurationError("Our encoder semantics assumes batch is always first!")
        except AttributeError:
            pass

    def get_input_dim(self) -> int:
        return self._module.input_size

    def get_output_dim(self) -> int:
        try:
            is_bidirectional = self._module.bidirectional
        except AttributeError:
            is_bidirectional = False
        return self._module.hidden_size * (2 if is_bidirectional else 1)

    def forward(self,  # pylint: disable=arguments-differ
                inputs: torch.Tensor,
                mask: torch.Tensor,
                hidden_state: torch.Tensor = None) -> torch.Tensor:

        if mask is None:
            # If a mask isn't passed, there is no padding in the batch of instances, so we can just
            # return the last sequence output as the state.  This doesn't work in the case of
            # variable length sequences, as the last state for each element of the batch won't be
            # at the end of the max sequence length, so we have to use the state of the RNN below.
            return self._module(inputs, hidden_state)[0][:, -1, :]

        batch_size = mask.size(0)

        _, state, restoration_indices, = \
            self.sort_and_run_forward(self._module, inputs, mask, hidden_state)

        # Deal with the fact the LSTM state is a tuple of (state, memory).
        if isinstance(state, tuple):
            state = state[0]

        num_layers_times_directions, num_valid, encoding_dim = state.size()
        # Add back invalid rows.
        if num_valid < batch_size:
            # batch size is the second dimension here, because pytorch
            # returns RNN state as a tensor of shape (num_layers * num_directions,
            # batch_size, hidden_size)
            zeros = state.new_zeros(num_layers_times_directions,
                                    batch_size - num_valid,
                                    encoding_dim)
            state = torch.cat([state, zeros], 1)

        # Restore the original indices and return the final state of the
        # top layer. Pytorch's recurrent layers return state in the form
        # (num_layers * num_directions, batch_size, hidden_size) regardless
        # of the 'batch_first' flag, so we transpose, extract the relevant
        # layer state (both forward and backward if using bidirectional layers)
        # and return them as a single (batch_size, self.get_output_dim()) tensor.

        # now of shape: (batch_size, num_layers * num_directions, hidden_size).
        unsorted_state = state.transpose(0, 1).index_select(0, restoration_indices)

        # Extract the last hidden vector, including both forward and backward states
        # if the cell is bidirectional. Then reshape by concatenation (in the case
        # we have bidirectional states) or just squash the 1st dimension in the non-
        # bidirectional case. Return tensor has shape (batch_size, hidden_size * num_directions).
        try:
            last_state_index = 2 if self._module.bidirectional else 1
        except AttributeError:
            last_state_index = 1
        last_layer_state = unsorted_state[:, -last_state_index:, :]
        return last_layer_state.contiguous().view([-1, self.get_output_dim()])



# pylint: disable=no-self-use,invalid-name

from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor


class TestDialogQAPredictor(AllenNlpTestCase):
    def test_uses_named_inputs(self):
        inputs = {"paragraphs": [{"qas": [{"followup": "y", "yesno": "x", "question": "When was the first one?",
                                           "answers": [{"answer_start": 0, "text": "One time"}], "id": "C_q#0"},
                                          {"followup": "n", "yesno": "x", "question": "What were you doing?",
                                           "answers": [{"answer_start": 15, "text": "writing a"}], "id": "C_q#1"},
                                          {"followup": "m", "yesno": "y", "question": "How often?",
                                           "answers": [{"answer_start": 4, "text": "time I"}], "id": "C_q#2"}],
                                  "context": "One time I was writing a unit test,\
                                   and it succeeded on the first attempt."}]}

        archive = load_archive(self.FIXTURES_ROOT / 'dialog_qa' / 'serialization' / 'model.tar.gz')
        predictor = Predictor.from_archive(archive, 'dialog_qa')

        result = predictor.predict_json(inputs)

        best_span_str_list = result.get("best_span_str")
        for best_span_str in best_span_str_list:
            assert isinstance(best_span_str, str)
            assert best_span_str != ""

    def test_batch_prediction(self):
        inputs = [{"paragraphs": [{"qas": [{"followup": "y", "yesno": "x", "question": "When was the first one?",
                                            "answers": [{"answer_start": 0, "text": "One time"}], "id": "C_q#0"},
                                           {"followup": "n", "yesno": "x", "question": "What were you doing?",
                                            "answers": [{"answer_start": 15, "text": "writing a"}], "id": "C_q#1"},
                                           {"followup": "m", "yesno": "y", "question": "How often?",
                                            "answers": [{"answer_start": 4, "text": "time I"}], "id": "C_q#2"}],
                                   "context": "One time I was writing a unit test,\
                                    and it succeeded on the first attempt."}]},
                  {"paragraphs": [{"qas": [{"followup": "y", "yesno": "x", "question": "When was the first one?",
                                            "answers": [{"answer_start": 0, "text": "One time"}], "id": "C_q#0"},
                                           {"followup": "n", "yesno": "x", "question": "What were you doing?",
                                            "answers": [{"answer_start": 15, "text": "writing a"}], "id": "C_q#1"},
                                           {"followup": "m", "yesno": "y", "question": "How often?",
                                            "answers": [{"answer_start": 4, "text": "time I"}], "id": "C_q#2"}],
                                   "context": "One time I was writing a unit test,\
                                    and it succeeded on the first attempt."}]}]

        archive = load_archive(self.FIXTURES_ROOT / 'dialog_qa' / 'serialization' / 'model.tar.gz')
        predictor = Predictor.from_archive(archive, 'dialog_qa')

        results = predictor.predict_batch_json(inputs)
        assert len(results) == 2



"""
Base class for subcommands under ``allennlp.run``.
"""
import argparse

class Subcommand:
    """
    An abstract class representing subcommands for allennlp.run.
    If you wanted to (for example) create your own custom `special-evaluate` command to use like

    ``allennlp special-evaluate ...``

    you would create a ``Subcommand`` subclass and then pass it as an override to
    :func:`~allennlp.commands.main` .
    """
    def add_subparser(self, name: str, parser: argparse._SubParsersAction) -> argparse.ArgumentParser:
        # pylint: disable=protected-access
        raise NotImplementedError



from typing import Dict, List
import itertools

from overrides import overrides

from allennlp.common.util import pad_sequence_to_length
from allennlp.data.vocabulary import Vocabulary
from allennlp.data.tokenizers.token import Token
from allennlp.data.token_indexers.token_indexer import TokenIndexer


@TokenIndexer.register("single_id")
class SingleIdTokenIndexer(TokenIndexer[int]):
    """
    This :class:`TokenIndexer` represents tokens as single integers.

    Parameters
    ----------
    namespace : ``str``, optional (default=``tokens``)
        We will use this namespace in the :class:`Vocabulary` to map strings to indices.
    lowercase_tokens : ``bool``, optional (default=``False``)
        If ``True``, we will call ``token.lower()`` before getting an index for the token from the
        vocabulary.
    start_tokens : ``List[str]``, optional (default=``None``)
        These are prepended to the tokens provided to ``tokens_to_indices``.
    end_tokens : ``List[str]``, optional (default=``None``)
        These are appended to the tokens provided to ``tokens_to_indices``.
    """
    # pylint: disable=no-self-use
    def __init__(self,
                 namespace: str = 'tokens',
                 lowercase_tokens: bool = False,
                 start_tokens: List[str] = None,
                 end_tokens: List[str] = None) -> None:
        self.namespace = namespace
        self.lowercase_tokens = lowercase_tokens

        self._start_tokens = [Token(st) for st in (start_tokens or [])]
        self._end_tokens = [Token(et) for et in (end_tokens or [])]

    @overrides
    def count_vocab_items(self, token: Token, counter: Dict[str, Dict[str, int]]):
        # If `text_id` is set on the token (e.g., if we're using some kind of hash-based word
        # encoding), we will not be using the vocab for this token.
        if getattr(token, 'text_id', None) is None:
            text = token.text
            if self.lowercase_tokens:
                text = text.lower()
            counter[self.namespace][text] += 1

    @overrides
    def tokens_to_indices(self,
                          tokens: List[Token],
                          vocabulary: Vocabulary,
                          index_name: str) -> Dict[str, List[int]]:
        indices: List[int] = []

        for token in itertools.chain(self._start_tokens, tokens, self._end_tokens):
            if getattr(token, 'text_id', None) is not None:
                # `text_id` being set on the token means that we aren't using the vocab, we just use
                # this id instead.
                indices.append(token.text_id)
            else:
                text = token.text
                if self.lowercase_tokens:
                    text = text.lower()
                indices.append(vocabulary.get_token_index(text, self.namespace))

        return {index_name: indices}

    @overrides
    def get_padding_token(self) -> int:
        return 0

    @overrides
    def get_padding_lengths(self, token: int) -> Dict[str, int]:  # pylint: disable=unused-argument
        return {}

    @overrides
    def pad_token_sequence(self,
                           tokens: Dict[str, List[int]],
                           desired_num_tokens: Dict[str, int],
                           padding_lengths: Dict[str, int]) -> Dict[str, List[int]]:  # pylint: disable=unused-argument
        return {key: pad_sequence_to_length(val, desired_num_tokens[key])
                for key, val in tokens.items()}



"""
A ``TextField`` represents a string of text, the kind that you might want to represent with
standard word vectors, or pass through an LSTM.
"""
from typing import Dict, List, Optional, Iterator
import textwrap

from overrides import overrides
from spacy.tokens import Token as SpacyToken
import torch

from allennlp.common.checks import ConfigurationError
from allennlp.data.fields.sequence_field import SequenceField
from allennlp.data.tokenizers.token import Token
from allennlp.data.token_indexers.token_indexer import TokenIndexer, TokenType
from allennlp.data.vocabulary import Vocabulary
from allennlp.nn import util

TokenList = List[TokenType]  # pylint: disable=invalid-name


class TextField(SequenceField[Dict[str, torch.Tensor]]):
    """
    This ``Field`` represents a list of string tokens.  Before constructing this object, you need
    to tokenize raw strings using a :class:`~allennlp.data.tokenizers.tokenizer.Tokenizer`.

    Because string tokens can be represented as indexed arrays in a number of ways, we also take a
    dictionary of :class:`~allennlp.data.token_indexers.token_indexer.TokenIndexer`
    objects that will be used to convert the tokens into indices.
    Each ``TokenIndexer`` could represent each token as a single ID, or a list of character IDs, or
    something else.

    This field will get converted into a dictionary of arrays, one for each ``TokenIndexer``.  A
    ``SingleIdTokenIndexer`` produces an array of shape (num_tokens,), while a
    ``TokenCharactersIndexer`` produces an array of shape (num_tokens, num_characters).
    """
    def __init__(self, tokens: List[Token], token_indexers: Dict[str, TokenIndexer]) -> None:
        self.tokens = tokens
        self._token_indexers = token_indexers
        self._indexed_tokens: Optional[Dict[str, TokenList]] = None
        self._indexer_name_to_indexed_token: Optional[Dict[str, List[str]]] = None

        if not all([isinstance(x, (Token, SpacyToken)) for x in tokens]):
            raise ConfigurationError("TextFields must be passed Tokens. "
                                     "Found: {} with types {}.".format(tokens, [type(x) for x in tokens]))

    # Sequence[Token] methods
    def __iter__(self) -> Iterator[Token]:
        return iter(self.tokens)

    def __getitem__(self, idx: int) -> Token:
        return self.tokens[idx]

    def __len__(self) -> int:
        return len(self.tokens)

    @overrides
    def count_vocab_items(self, counter: Dict[str, Dict[str, int]]):
        for indexer in self._token_indexers.values():
            for token in self.tokens:
                indexer.count_vocab_items(token, counter)

    @overrides
    def index(self, vocab: Vocabulary):
        token_arrays: Dict[str, TokenList] = {}
        indexer_name_to_indexed_token: Dict[str, List[str]] = {}
        for indexer_name, indexer in self._token_indexers.items():
            token_indices = indexer.tokens_to_indices(self.tokens, vocab, indexer_name)
            token_arrays.update(token_indices)
            indexer_name_to_indexed_token[indexer_name] = list(token_indices.keys())
        self._indexed_tokens = token_arrays
        self._indexer_name_to_indexed_token = indexer_name_to_indexed_token

    @overrides
    def get_padding_lengths(self) -> Dict[str, int]:
        """
        The ``TextField`` has a list of ``Tokens``, and each ``Token`` gets converted into arrays by
        (potentially) several ``TokenIndexers``.  This method gets the max length (over tokens)
        associated with each of these arrays.
        """
        # Our basic outline: we will iterate over `TokenIndexers`, and aggregate lengths over tokens
        # for each indexer separately.  Then we will combine the results for each indexer into a single
        # dictionary, resolving any (unlikely) key conflicts by taking a max.
        lengths = []
        if self._indexed_tokens is None:
            raise ConfigurationError("You must call .index(vocabulary) on a "
                                     "field before determining padding lengths.")

        # Each indexer can return a different sequence length, and for indexers that return
        # multiple arrays each can have a different length.  We'll keep track of them here.
        for indexer_name, indexer in self._token_indexers.items():
            indexer_lengths = {}

            for indexed_tokens_key in self._indexer_name_to_indexed_token[indexer_name]:
                # This is a list of dicts, one for each token in the field.
                token_lengths = [indexer.get_padding_lengths(token)
                                 for token in self._indexed_tokens[indexed_tokens_key]]
            if not token_lengths:
                # This is a padding edge case and occurs when we want to pad a ListField of
                # TextFields. In order to pad the list field, we need to be able to have an
                # _empty_ TextField, but if this is the case, token_lengths will be an empty
                # list, so we add the default empty padding dictionary to the list instead.
                token_lengths = [{}]
            # Iterate over the keys and find the maximum token length.
            # It's fine to iterate over the keys of the first token since all tokens have the same keys.
            for key in token_lengths[0]:
                indexer_lengths[key] = max(x[key] if key in x else 0 for x in token_lengths)
            lengths.append(indexer_lengths)

        padding_lengths = {}
        num_tokens = set()
        for indexer_name, token_list in self._indexed_tokens.items():
            padding_lengths[f"{indexer_name}_length"] = len(token_list)
            num_tokens.add(len(token_list))

        # We don't actually use this for padding anywhere, but we used to.  We add this key back in
        # so that older configs still work if they sorted by this key in a BucketIterator.  Taking
        # the max of all of these should be fine for that purpose.
        padding_lengths['num_tokens'] = max(num_tokens)

        # Get all keys which have been used for padding for each indexer and take the max if there are duplicates.
        padding_keys = {key for d in lengths for key in d.keys()}
        for padding_key in padding_keys:
            padding_lengths[padding_key] = max(x[padding_key] if padding_key in x else 0 for x in lengths)
        return padding_lengths

    @overrides
    def sequence_length(self) -> int:
        return len(self.tokens)

    @overrides
    def as_tensor(self, padding_lengths: Dict[str, int]) -> Dict[str, torch.Tensor]:
        tensors = {}
        for indexer_name, indexer in self._token_indexers.items():
            desired_num_tokens = {indexed_tokens_key: padding_lengths[f"{indexed_tokens_key}_length"]
                                  for indexed_tokens_key in self._indexer_name_to_indexed_token[indexer_name]}
            indices_to_pad = {indexed_tokens_key: self._indexed_tokens[indexed_tokens_key]
                              for indexed_tokens_key in self._indexer_name_to_indexed_token[indexer_name]}
            padded_array = indexer.pad_token_sequence(indices_to_pad,
                                                      desired_num_tokens, padding_lengths)
            # We use the key of the indexer to recognise what the tensor corresponds to within the
            # field (i.e. the result of word indexing, or the result of character indexing, for
            # example).
            # TODO(mattg): we might someday have a TokenIndexer that needs to use something other
            # than a LongTensor here, and it's not clear how to signal that.  Maybe we'll need to
            # add a class method to TokenIndexer to tell us the type?  But we can worry about that
            # when there's a compelling use case for it.
            indexer_tensors = {key: torch.LongTensor(array) for key, array in padded_array.items()}
            tensors.update(indexer_tensors)
        return tensors

    @overrides
    def empty_field(self):
        # pylint: disable=protected-access
        text_field = TextField([], self._token_indexers)
        text_field._indexed_tokens = {}
        text_field._indexer_name_to_indexed_token = {}
        for indexer_name, indexer in self._token_indexers.items():
            array_keys = indexer.get_keys(indexer_name)
            for key in array_keys:
                text_field._indexed_tokens[key] = []
            text_field._indexer_name_to_indexed_token[indexer_name] = array_keys
        return text_field

    @overrides
    def batch_tensors(self, tensor_list: List[Dict[str, torch.Tensor]]) -> Dict[str, torch.Tensor]:
        # pylint: disable=no-self-use
        # This is creating a dict of {token_indexer_key: batch_tensor} for each token indexer used
        # to index this field.
        return util.batch_tensor_dicts(tensor_list)

    def __str__(self) -> str:
        indexers = {name: indexer.__class__.__name__ for name, indexer in self._token_indexers.items()}

        # Double tab to indent under the header.
        formatted_text = "".join(["\t\t" + text + "\n"
                                  for text in textwrap.wrap(repr(self.tokens), 100)])
        return f"TextField of length {self.sequence_length()} with " \
               f"text: \n {formatted_text} \t\tand TokenIndexers : {indexers}"



from allennlp.common.testing import ModelTestCase, AllenNlpTestCase


class TestBasicClassifier(ModelTestCase):

    def test_seq2vec_clf_can_train_save_and_load(self):
        data_directory = AllenNlpTestCase.FIXTURES_ROOT / "data" / "text_classification_json"
        self.set_up_model(AllenNlpTestCase.FIXTURES_ROOT / 'basic_classifier' / 'experiment_seq2vec.jsonnet',
                          data_directory / "imdb_corpus.jsonl")
        self.ensure_model_can_train_save_and_load(self.param_file)

    def test_seq2seq_clf_can_train_save_and_load(self):
        data_directory = AllenNlpTestCase.FIXTURES_ROOT / "data" / "text_classification_json"
        self.set_up_model(AllenNlpTestCase.FIXTURES_ROOT / 'basic_classifier' / 'experiment_seq2seq.jsonnet',
                          data_directory / "imdb_corpus.jsonl")
        self.ensure_model_can_train_save_and_load(self.param_file)



# pylint: disable=invalid-name,no-self-use,protected-access
import numpy
from numpy.testing import assert_almost_equal
import torch
from torch.nn import Parameter

from allennlp.common import Params
from allennlp.common.testing.test_case import AllenNlpTestCase
from allennlp.modules.matrix_attention import LinearMatrixAttention
from allennlp.modules.matrix_attention.matrix_attention import MatrixAttention


class TestLinearMatrixAttention(AllenNlpTestCase):

    def test_can_init_dot(self):
        legacy_attention = MatrixAttention.from_params(Params({"type": "linear",
                                                               "tensor_1_dim": 3,
                                                               "tensor_2_dim": 3}))
        isinstance(legacy_attention, LinearMatrixAttention)

    def test_linear_similarity(self):
        linear = LinearMatrixAttention(3, 3)
        linear._weight_vector = Parameter(torch.FloatTensor([-.3, .5, 2.0, -1.0, 1, 1]))
        linear._bias = Parameter(torch.FloatTensor([.1]))
        output = linear(torch.FloatTensor([[[0, 0, 0], [4, 5, 6]], [[-7, -8, -9], [10, 11, 12]]]),
                        torch.FloatTensor([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]]))

        assert_almost_equal(output.data.numpy(), numpy.array([[[4.1000, 7.1000], [17.4000, 20.4000]],
                                                              [[-9.8000, -6.8000], [36.6000, 39.6000]]]),
                            decimal=2)

    def test_bidaf_trilinear_similarity(self):
        linear = LinearMatrixAttention(2, 2, combination='x,y,x*y')
        linear._weight_vector = Parameter(torch.FloatTensor([-.3, .5, 2.0, -1.0, 1, 1]))
        linear._bias = Parameter(torch.FloatTensor([.0]))
        output = linear(torch.FloatTensor([[[0, 0], [4, 5]], [[-7, -8], [10, 11]]]),
                        torch.FloatTensor([[[1, 2], [4, 5]], [[7, 8], [10, 11]]]))

        assert_almost_equal(output.data.numpy(),
                            numpy.array([[[0 + 0 + 2 + -2 + 0 + 0,
                                           0 + 0 + 8 + -5 + 0 + 0],
                                          [-1.2 + 2.5 + 2 + -2 + 4 + 10,
                                           -1.2 + 2.5 + 8 + -5 + 16 + 25]],
                                         [[2.1 + -4 + 14 + -8 + -49 + -64,
                                           2.1 + -4 + 20 + -11 + -70 + -88],
                                          [-3 + 5.5 + 14 + -8 + 70 + 88,
                                           -3 + 5.5 + 20 + -11 + 100 + 121]]]),
                            decimal=2)



from overrides import overrides
import torch
from torch.nn.parameter import Parameter

from allennlp.modules.matrix_attention.matrix_attention import MatrixAttention
from allennlp.nn import Activation


@MatrixAttention.register("bilinear")
class BilinearMatrixAttention(MatrixAttention):
    """
    Computes attention between two matrices using a bilinear attention function.  This function has
    a matrix of weights ``W`` and a bias ``b``, and the similarity between the two matrices ``X``
    and ``Y`` is computed as ``X W Y^T + b``.

    Parameters
    ----------
    matrix_1_dim : ``int``
        The dimension of the matrix ``X``, described above.  This is ``X.size()[-1]`` - the length
        of the vector that will go into the similarity computation.  We need this so we can build
        the weight matrix correctly.
    matrix_2_dim : ``int``
        The dimension of the matrix ``Y``, described above.  This is ``Y.size()[-1]`` - the length
        of the vector that will go into the similarity computation.  We need this so we can build
        the weight matrix correctly.
    activation : ``Activation``, optional (default=linear (i.e. no activation))
        An activation function applied after the ``X W Y^T + b`` calculation.  Default is no
        activation.
    use_input_biases : ``bool``, optional (default = False)
        If True, we add biases to the inputs such that the final computation
        is equivalent to the original bilinear matrix multiplication plus a
        projection of both inputs.
    label_dim : ``int``, optional (default = 1)
        The number of output classes. Typically in an attention setting this will be one,
        but this parameter allows this class to function as an equivalent to ``torch.nn.Bilinear``
        for matrices, rather than vectors.
    """
    def __init__(self,
                 matrix_1_dim: int,
                 matrix_2_dim: int,
                 activation: Activation = None,
                 use_input_biases: bool = False,
                 label_dim: int = 1) -> None:
        super().__init__()
        if use_input_biases:
            matrix_1_dim += 1
            matrix_2_dim += 1

        if label_dim == 1:
            self._weight_matrix = Parameter(torch.Tensor(matrix_1_dim, matrix_2_dim))
        else:
            self._weight_matrix = Parameter(torch.Tensor(label_dim, matrix_1_dim, matrix_2_dim))

        self._bias = Parameter(torch.Tensor(1))
        self._activation = activation or Activation.by_name('linear')()
        self._use_input_biases = use_input_biases
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self._weight_matrix)
        self._bias.data.fill_(0)

    @overrides
    def forward(self, matrix_1: torch.Tensor, matrix_2: torch.Tensor) -> torch.Tensor:

        if self._use_input_biases:
            bias1 = matrix_1.new_ones(matrix_1.size()[:-1] + (1,))
            bias2 = matrix_2.new_ones(matrix_2.size()[:-1] + (1,))

            matrix_1 = torch.cat([matrix_1, bias1], -1)
            matrix_2 = torch.cat([matrix_2, bias2], -1)

        weight = self._weight_matrix
        if weight.dim() == 2:
            weight = weight.unsqueeze(0)
        intermediate = torch.matmul(matrix_1.unsqueeze(1), weight)
        final = torch.matmul(intermediate, matrix_2.unsqueeze(1).transpose(2, 3))
        return self._activation(final.squeeze(1) + self._bias)



"""
AllenNLP uses most
`PyTorch learning rate schedulers <http://pytorch.org/docs/master/optim.html#how-to-adjust-learning-rate>`_,
with a thin wrapper to allow registering them and instantiating them ``from_params``.

The available learning rate schedulers from PyTorch are

* `"step" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.StepLR>`_
* `"multi_step" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.MultiStepLR>`_
* `"exponential" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ExponentialLR>`_
* `"reduce_on_plateau" <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau>`_

In addition, AllenNLP also provides `cosine with restarts <https://arxiv.org/abs/1608.03983>`_,
a Noam schedule, and a slanted triangular schedule, which are registered as
"cosine", "noam", and "slanted_triangular", respectively.
"""

from allennlp.training.learning_rate_schedulers.learning_rate_scheduler import LearningRateScheduler
from allennlp.training.learning_rate_schedulers.cosine import CosineWithRestarts
from allennlp.training.learning_rate_schedulers.noam import NoamLR
from allennlp.training.learning_rate_schedulers.slanted_triangular import SlantedTriangular



# pylint: disable=invalid-name,no-self-use,protected-access
import torch
from numpy.testing import assert_almost_equal
import numpy

from allennlp.common import Params
from allennlp.common.testing.test_case import AllenNlpTestCase
from allennlp.modules.matrix_attention import CosineMatrixAttention
from allennlp.modules.matrix_attention.matrix_attention import MatrixAttention


class TestCosineMatrixAttention(AllenNlpTestCase):

    def test_can_init_cosine(self):
        legacy_attention = MatrixAttention.from_params(Params({"type": "cosine"}))
        isinstance(legacy_attention, CosineMatrixAttention)

    def test_cosine_similarity(self):
        # example use case: a batch of size 2.
        # With a time element component (e.g. sentences of length 2) each word is a vector of length 3.
        # It is comparing this with another input of the same type
        output = CosineMatrixAttention()(torch.FloatTensor([[[0, 0, 0], [4, 5, 6]], [[-7, -8, -9], [10, 11, 12]]]),
                                         torch.FloatTensor([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]]))

        # For the first batch there is
        #       no correlation between the first words of the input matrix
        #       but perfect correlation for the second word
        # For the second batch there is
        #     negative correlation for the first words
        #     correlation for the second word
        assert_almost_equal(output.numpy(), numpy.array([[[0, 0], [.97, 1]], [[-1, -0.99], [0.99, 1]]]),
                            decimal=2)



# pylint: disable=no-self-use,invalid-name
from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor


class TestSrlPredictor(AllenNlpTestCase):
    def test_uses_named_inputs(self):
        inputs = {
                "sentence": "The squirrel wrote a unit test to make sure its nuts worked as designed."
        }

        archive = load_archive(self.FIXTURES_ROOT / 'srl' / 'serialization' / 'model.tar.gz')
        predictor = Predictor.from_archive(archive, 'semantic-role-labeling')

        result_json = predictor.predict_json(inputs)
        self.assert_predict_result(result_json)

        words = ["The", "squirrel", "wrote", "a", "unit", "test",
                 "to", "make", "sure", "its", "nuts", "worked", "as", "designed", "."]

        result_words = predictor.predict_tokenized(words)
        self.assert_predict_result(result_words)

    @staticmethod
    def assert_predict_result(result):
        words = result.get("words")
        assert words == ["The", "squirrel", "wrote", "a", "unit", "test",
                         "to", "make", "sure", "its", "nuts", "worked", "as", "designed", "."]

        num_words = len(words)
        verbs = result.get("verbs")
        assert verbs is not None
        assert isinstance(verbs, list)
        assert any(v["verb"] == "wrote" for v in verbs)
        assert any(v["verb"] == "make" for v in verbs)
        assert any(v["verb"] == "worked" for v in verbs)
        for verb in verbs:
            tags = verb.get("tags")
            assert tags is not None
            assert isinstance(tags, list)
            assert all(isinstance(tag, str) for tag in tags)
            assert len(tags) == num_words

    def test_batch_prediction(self):
        inputs = {
                "sentence": "The squirrel wrote a unit test to make sure its nuts worked as designed."
        }
        archive = load_archive(self.FIXTURES_ROOT / 'srl' / 'serialization' / 'model.tar.gz')
        predictor = Predictor.from_archive(archive, 'semantic-role-labeling')
        result = predictor.predict_batch_json([inputs, inputs])
        assert result[0] == result[1]

    def test_prediction_with_no_verbs(self):

        input1 = {"sentence": "Blah no verb sentence."}
        archive = load_archive(self.FIXTURES_ROOT / 'srl' / 'serialization' / 'model.tar.gz')
        predictor = Predictor.from_archive(archive, 'semantic-role-labeling')
        result = predictor.predict_json(input1)
        assert result == {'words': ['Blah', 'no', 'verb', 'sentence', '.'], 'verbs': []}

        input2 = {"sentence": "This sentence has a verb."}
        results = predictor.predict_batch_json([input1, input2])
        assert results[0] == {'words': ['Blah', 'no', 'verb', 'sentence', '.'], 'verbs': []}
        assert results[1] == {'words': ['This', 'sentence', 'has', 'a', 'verb', '.'],
                              'verbs': [{'verb': 'has', 'description': 'This sentence has a verb .',
                                         'tags': ['O', 'O', 'O', 'O', 'O', 'O']}]}



# pylint: disable=no-self-use,invalid-name
from typing import Dict

import pytest

from allennlp.common.configuration import configure, Config, BASE_CONFIG, json_annotation, choices
from allennlp.common.testing import AllenNlpTestCase
from allennlp.nn.activations import Activation


class TestConfiguration(AllenNlpTestCase):
    def test_configure_top_level(self):
        config = configure()

        assert config == BASE_CONFIG

    def test_abstract_base_class(self):
        config = choices('allennlp.data.dataset_readers.dataset_reader.DatasetReader')

        assert isinstance(config, list)
        assert 'allennlp.data.dataset_readers.snli.SnliReader' in config

    def test_specific_subclass(self):
        config = configure('allennlp.data.dataset_readers.semantic_role_labeling.SrlReader')
        assert isinstance(config, Config)

        items = {item.name: item for item in config.items}

        assert len(items) == 3

        assert 'token_indexers' in items
        token_indexers = items['token_indexers']
        assert token_indexers.default_value is None

        assert 'domain_identifier' in items
        domain_identifier = items['domain_identifier']
        assert domain_identifier.annotation == str
        assert domain_identifier.default_value is None

        assert 'lazy' in items
        lazy = items['lazy']
        assert lazy.annotation == bool
        assert not lazy.default_value

    def test_errors(self):
        with pytest.raises(ModuleNotFoundError):
            configure('allennlp.non_existent_module.SomeClass')

        with pytest.raises(AttributeError):
            configure('allennlp.data.dataset_readers.NonExistentDatasetReader')

    def test_vocab_workaround(self):
        config = configure('allennlp.data.vocabulary.Vocabulary')
        assert isinstance(config, Config)

        items = {item.name: item for item in config.items}

        assert len(items) == 9
        assert "directory_path" in items
        assert "max_vocab_size" in items

    def test_activation_workaround(self):
        annotation = Dict[str, Activation]
        ja = json_annotation(annotation)

        assert ja == {
                "origin": "Dict",
                "args": [
                        {"origin": "str"},
                        {"origin": "str"}
                ]
        }



# pylint: disable=no-self-use,invalid-name
from flaky import flaky
import pytest
import numpy
from numpy.testing import assert_almost_equal

from allennlp.common import Params
from allennlp.common.checks import ConfigurationError
from allennlp.common.testing import ModelTestCase
from allennlp.models import DecomposableAttention, Model


class TestDecomposableAttention(ModelTestCase):
    def setUp(self):
        super(TestDecomposableAttention, self).setUp()
        self.set_up_model(self.FIXTURES_ROOT / 'decomposable_attention' / 'experiment.json',
                          self.FIXTURES_ROOT / 'data' / 'snli.jsonl')

    def test_forward_pass_runs_correctly(self):
        training_tensors = self.dataset.as_tensor_dict()
        output_dict = self.model(**training_tensors)
        assert_almost_equal(numpy.sum(output_dict["label_probs"][0].data.numpy(), -1), 1, decimal=6)

    @flaky
    def test_model_can_train_save_and_load(self):
        self.ensure_model_can_train_save_and_load(self.param_file)

    @flaky
    def test_batch_predictions_are_consistent(self):
        self.ensure_batch_predictions_are_consistent()

    def test_model_load(self):
        params = Params.from_file(self.FIXTURES_ROOT / 'decomposable_attention' / 'experiment.json')
        model = Model.load(params, serialization_dir=self.FIXTURES_ROOT /
                           'decomposable_attention' / 'serialization')

        assert isinstance(model, DecomposableAttention)

    def test_mismatched_dimensions_raise_configuration_errors(self):
        params = Params.from_file(self.param_file)
        # Make the input_dim to the first feedforward_layer wrong - it should be 2.
        params["model"]["attend_feedforward"]["input_dim"] = 10
        with pytest.raises(ConfigurationError):
            Model.from_params(vocab=self.vocab, params=params.pop("model"))

        params = Params.from_file(self.param_file)
        # Make the projection output_dim of the last layer wrong - it should be
        # 3, equal to the number of classes.
        params["model"]["aggregate_feedforward"]["output_dim"] = 10
        with pytest.raises(ConfigurationError):
            Model.from_params(vocab=self.vocab, params=params.pop("model"))



import torch

from allennlp.common import Registrable

class SimilarityFunction(torch.nn.Module, Registrable):
    """
    A ``SimilarityFunction`` takes a pair of tensors with the same shape, and computes a similarity
    function on the vectors in the last dimension.  For example, the tensors might both have shape
    `(batch_size, sentence_length, embedding_dim)`, and we will compute some function of the two
    vectors of length `embedding_dim` for each position `(batch_size, sentence_length)`, returning a
    tensor of shape `(batch_size, sentence_length)`.

    The similarity function could be as simple as a dot product, or it could be a more complex,
    parameterized function.

    If you want to compute a similarity between tensors of different sizes, you need to first tile
    them in the appropriate dimensions to make them the same before you can use these functions.
    The :class:`~allennlp.modules.Attention` and :class:`~allennlp.modules.MatrixAttention` modules
    do this.
    """
    default_implementation = 'dot_product'

    def forward(self, tensor_1: torch.Tensor, tensor_2: torch.Tensor) -> torch.Tensor:
        # pylint: disable=arguments-differ
        """
        Takes two tensors of the same shape, such as ``(batch_size, length_1, length_2,
        embedding_dim)``.  Computes a (possibly parameterized) similarity on the final dimension
        and returns a tensor with one less dimension, such as ``(batch_size, length_1, length_2)``.
        """
        raise NotImplementedError



"""
Defines all the types in the QuaRel domain.
"""
from allennlp.semparse.type_declarations.type_declaration import ComplexType, NamedBasicType, NameMapper

class QuarelTypeDeclaration:
    def __init__(self, syntax: str) -> None:

        self.name_mapper = NameMapper()

        num_type = NamedBasicType("NUM")
        attr_type = NamedBasicType("ATTR")
        rdir_type = NamedBasicType("RDIR")
        world_type = NamedBasicType("WORLD")
        var_type = NamedBasicType("VAR")

        self.basic_types = {num_type, attr_type, rdir_type, world_type, var_type}

        if syntax == "quarel_friction":
            # attributes: <<QDIR, <WORLD, ATTR>>
            attr_function_type = ComplexType(rdir_type,
                                             ComplexType(world_type, attr_type))

            and_function_type = ComplexType(attr_type, ComplexType(attr_type, attr_type))

            # infer: <ATTR, <ATTR, <ATTR, NUM>>>
            infer_function_type = ComplexType(attr_type,
                                              ComplexType(attr_type,
                                                          ComplexType(attr_type, num_type)))
            self.name_mapper.map_name_with_signature("infer", infer_function_type)
            # Attributes
            self.name_mapper.map_name_with_signature("friction", attr_function_type)
            self.name_mapper.map_name_with_signature("smoothness", attr_function_type)
            self.name_mapper.map_name_with_signature("speed", attr_function_type)
            self.name_mapper.map_name_with_signature("heat", attr_function_type)
            self.name_mapper.map_name_with_signature("distance", attr_function_type)

            # For simplicity we treat "high" and "low" as directions as well
            self.name_mapper.map_name_with_signature("high", rdir_type)
            self.name_mapper.map_name_with_signature("low", rdir_type)
            self.name_mapper.map_name_with_signature("and", and_function_type)

            self.curried_functions = {
                    attr_function_type: 2,
                    infer_function_type: 3,
                    and_function_type: 2
            }
        elif syntax == "quarel_v1_attr_entities" or syntax == "quarel_friction_attr_entities":
            # attributes: <<QDIR, <WORLD, ATTR>>
            attr_function_type = ComplexType(rdir_type,
                                             ComplexType(world_type, attr_type))

            and_function_type = ComplexType(attr_type, ComplexType(attr_type, attr_type))

            # infer: <ATTR, <ATTR, <ATTR, NUM>>>
            infer_function_type = ComplexType(attr_type,
                                              ComplexType(attr_type,
                                                          ComplexType(attr_type, num_type)))
            self.name_mapper.map_name_with_signature("infer", infer_function_type)
            # TODO: Remove this?
            self.name_mapper.map_name_with_signature("placeholder", attr_function_type)

            # For simplicity we treat "high" and "low" as directions as well
            self.name_mapper.map_name_with_signature("high", rdir_type)
            self.name_mapper.map_name_with_signature("low", rdir_type)
            self.name_mapper.map_name_with_signature("and", and_function_type)

            self.curried_functions = {
                    attr_function_type: 2,
                    infer_function_type: 3,
                    and_function_type: 2
            }

        elif syntax == "quarel_v1":
            # attributes: <<QDIR, <WORLD, ATTR>>
            attr_function_type = ComplexType(rdir_type,
                                             ComplexType(world_type, attr_type))

            and_function_type = ComplexType(attr_type, ComplexType(attr_type, attr_type))

            # infer: <ATTR, <ATTR, <ATTR, NUM>>>
            infer_function_type = ComplexType(attr_type,
                                              ComplexType(attr_type,
                                                          ComplexType(attr_type, num_type)))
            self.name_mapper.map_name_with_signature("infer", infer_function_type)
            # Attributes
            self.name_mapper.map_name_with_signature("friction", attr_function_type)
            self.name_mapper.map_name_with_signature("smoothness", attr_function_type)
            self.name_mapper.map_name_with_signature("speed", attr_function_type)
            self.name_mapper.map_name_with_signature("heat", attr_function_type)
            self.name_mapper.map_name_with_signature("distance", attr_function_type)
            self.name_mapper.map_name_with_signature("acceleration", attr_function_type)
            self.name_mapper.map_name_with_signature("amountSweat", attr_function_type)
            self.name_mapper.map_name_with_signature("apparentSize", attr_function_type)
            self.name_mapper.map_name_with_signature("breakability", attr_function_type)
            self.name_mapper.map_name_with_signature("brightness", attr_function_type)
            self.name_mapper.map_name_with_signature("exerciseIntensity", attr_function_type)
            self.name_mapper.map_name_with_signature("flexibility", attr_function_type)
            self.name_mapper.map_name_with_signature("gravity", attr_function_type)
            self.name_mapper.map_name_with_signature("loudness", attr_function_type)
            self.name_mapper.map_name_with_signature("mass", attr_function_type)
            self.name_mapper.map_name_with_signature("strength", attr_function_type)
            self.name_mapper.map_name_with_signature("thickness", attr_function_type)
            self.name_mapper.map_name_with_signature("time", attr_function_type)
            self.name_mapper.map_name_with_signature("weight", attr_function_type)

            # For simplicity we treat "high" and "low" as directions as well
            self.name_mapper.map_name_with_signature("high", rdir_type)
            self.name_mapper.map_name_with_signature("low", rdir_type)
            self.name_mapper.map_name_with_signature("and", and_function_type)

            self.curried_functions = {
                    attr_function_type: 2,
                    infer_function_type: 3,
                    and_function_type: 2
            }

        else:
            raise Exception(f"Unknown LF syntax specification: {syntax}")

        self.name_mapper.map_name_with_signature("higher", rdir_type)
        self.name_mapper.map_name_with_signature("lower", rdir_type)

        self.name_mapper.map_name_with_signature("world1", world_type)
        self.name_mapper.map_name_with_signature("world2", world_type)

        # Hack to expose types
        self.world_type = world_type
        self.attr_function_type = attr_function_type
        self.var_type = var_type

        self.starting_types = {num_type}



# pylint: disable=no-self-use,invalid-name
import pytest

from allennlp.common.testing.test_case import AllenNlpTestCase
from allennlp.data.fields import MetadataField


class TestMetadataField(AllenNlpTestCase):
    def test_mapping_works_with_dict(self):
        field = MetadataField({"a": 1, "b": [0]})

        assert "a" in field
        assert field["a"] == 1
        assert len(field) == 2

        keys = {k for k in field}
        assert keys == {"a", "b"}

        values = [v for v in field.values()]
        assert len(values) == 2
        assert 1 in values
        assert [0] in values

    def test_mapping_raises_with_non_dict(self):
        field = MetadataField(0)

        with pytest.raises(TypeError):
            _ = field[0]

        with pytest.raises(TypeError):
            _ = len(field)

        with pytest.raises(TypeError):
            _ = [x for x in field]



from typing import Dict
import argparse
import logging

from overrides import overrides

from allennlp import __version__
from allennlp.commands.configure import Configure
from allennlp.commands.elmo import Elmo
from allennlp.commands.evaluate import Evaluate
from allennlp.commands.fine_tune import FineTune
from allennlp.commands.make_vocab import MakeVocab
from allennlp.commands.predict import Predict
from allennlp.commands.dry_run import DryRun
from allennlp.commands.subcommand import Subcommand
from allennlp.commands.test_install import TestInstall
from allennlp.commands.find_learning_rate import FindLearningRate
from allennlp.commands.train import Train
from allennlp.common.util import import_submodules

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ArgumentParserWithDefaults(argparse.ArgumentParser):
    """
    Custom argument parser that will display the default value for an argument
    in the help message.
    """

    _action_defaults_to_ignore = {"help", "store_true", "store_false", "store_const"}

    @staticmethod
    def _is_empty_default(default):
        if default is None:
            return True
        if isinstance(default, (str, list, tuple, set)):
            return not bool(default)
        return False

    @overrides
    def add_argument(self, *args, **kwargs):
        # pylint: disable=arguments-differ
        # Add default value to the help message when the default is meaningful.
        default = kwargs.get("default")
        if kwargs.get("action") not in self._action_defaults_to_ignore and not self._is_empty_default(default):
            description = kwargs.get("help") or ""
            kwargs["help"] = f"{description} (default = {default})"
        super().add_argument(*args, **kwargs)


def main(prog: str = None,
         subcommand_overrides: Dict[str, Subcommand] = {}) -> None:
    """
    The :mod:`~allennlp.run` command only knows about the registered classes in the ``allennlp``
    codebase. In particular, once you start creating your own ``Model`` s and so forth, it won't
    work for them, unless you use the ``--include-package`` flag.
    """
    # pylint: disable=dangerous-default-value
    parser = ArgumentParserWithDefaults(description="Run AllenNLP", usage='%(prog)s', prog=prog)
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)

    subparsers = parser.add_subparsers(title='Commands', metavar='')

    subcommands = {
            # Default commands
            "configure": Configure(),
            "train": Train(),
            "evaluate": Evaluate(),
            "predict": Predict(),
            "make-vocab": MakeVocab(),
            "elmo": Elmo(),
            "fine-tune": FineTune(),
            "dry-run": DryRun(),
            "test-install": TestInstall(),
            "find-lr": FindLearningRate(),

            # Superseded by overrides
            **subcommand_overrides
    }

    for name, subcommand in subcommands.items():
        subparser = subcommand.add_subparser(name, subparsers)
        # configure doesn't need include-package because it imports
        # whatever classes it needs.
        if name != "configure":
            subparser.add_argument('--include-package',
                                   type=str,
                                   action='append',
                                   default=[],
                                   help='additional packages to include')

    args = parser.parse_args()

    # If a subparser is triggered, it adds its work as `args.func`.
    # So if no such attribute has been added, no subparser was triggered,
    # so give the user some help.
    if 'func' in dir(args):
        # Import any additional modules needed (to register custom classes).
        for package_name in getattr(args, 'include_package', ()):
            import_submodules(package_name)
        args.func(args)
    else:
        parser.print_help()



# pytest: disable=no-self-use,invalid-name
import shutil
import sys

import pytest

from allennlp.commands import main
from allennlp.commands.subcommand import Subcommand
from allennlp.common.checks import ConfigurationError
from allennlp.common.testing import AllenNlpTestCase

class TestMain(AllenNlpTestCase):
    def test_fails_on_unknown_command(self):
        sys.argv = ["bogus",         # command
                    "unknown_model", # model_name
                    "bogus file",    # input_file
                    "--output-file", "bogus out file",
                    "--silent"]

        with self.assertRaises(SystemExit) as cm:  # pylint: disable=invalid-name
            main()

        assert cm.exception.code == 2  # argparse code for incorrect usage

    def test_subcommand_overrides(self):
        def do_nothing(_):
            pass

        class FakeEvaluate(Subcommand):
            add_subparser_called = False

            def add_subparser(self, name, parser):
                subparser = parser.add_parser(name,
                                              description="fake",
                                              help="fake help")

                subparser.set_defaults(func=do_nothing)
                self.add_subparser_called = True

                return subparser

        fake_evaluate = FakeEvaluate()

        sys.argv = ["allennlp", "evaluate"]
        main(subcommand_overrides={"evaluate": fake_evaluate})

        assert fake_evaluate.add_subparser_called

    def test_other_modules(self):
        # Create a new package in a temporary dir
        packagedir = self.TEST_DIR / 'testpackage'
        packagedir.mkdir()  # pylint: disable=no-member
        (packagedir / '__init__.py').touch()  # pylint: disable=no-member

        # And add that directory to the path
        sys.path.insert(0, str(self.TEST_DIR))

        # Write out a duplicate model there, but registered under a different name.
        from allennlp.models import simple_tagger
        with open(simple_tagger.__file__) as model_file:
            code = model_file.read().replace("""@Model.register("simple_tagger")""",
                                             """@Model.register("duplicate-test-tagger")""")

        with open(packagedir / 'model.py', 'w') as new_model_file:
            new_model_file.write(code)

        # Copy fixture there too.
        shutil.copy(self.FIXTURES_ROOT / 'data' / 'sequence_tagging.tsv', self.TEST_DIR)
        data_path = str(self.TEST_DIR / 'sequence_tagging.tsv')

        # Write out config file
        config_path = self.TEST_DIR / 'config.json'
        config_json = """{
                "model": {
                        "type": "duplicate-test-tagger",
                        "text_field_embedder": {
                                "token_embedders": {
                                        "tokens": {
                                                "type": "embedding",
                                                "embedding_dim": 5
                                        }
                                }
                        },
                        "encoder": {
                                "type": "lstm",
                                "input_size": 5,
                                "hidden_size": 7,
                                "num_layers": 2
                        }
                },
                "dataset_reader": {"type": "sequence_tagging"},
                "train_data_path": "$$$",
                "validation_data_path": "$$$",
                "iterator": {"type": "basic", "batch_size": 2},
                "trainer": {
                        "num_epochs": 2,
                        "optimizer": "adam"
                }
            }""".replace('$$$', data_path)
        with open(config_path, 'w') as config_file:
            config_file.write(config_json)

        serialization_dir = self.TEST_DIR / 'serialization'

        # Run train with using the non-allennlp module.
        sys.argv = ["allennlp",
                    "train", str(config_path),
                    "-s", str(serialization_dir)]

        # Shouldn't be able to find the model.
        with pytest.raises(ConfigurationError):
            main()

        # Now add the --include-package flag and it should work.
        # We also need to add --recover since the output directory already exists.
        sys.argv.extend(["--recover", "--include-package", 'testpackage'])

        main()

        # Rewrite out config file, but change a value.
        with open(config_path, 'w') as new_config_file:
            new_config_file.write(config_json.replace('"num_epochs": 2,', '"num_epochs": 4,'))

        # This should fail because the config.json does not match that in the serialization directory.
        with pytest.raises(ConfigurationError):
            main()

        sys.path.remove(str(self.TEST_DIR))



# pylint: disable=no-self-use,invalid-name,line-too-long

from allennlp.data.dataset_readers import TemplateText2SqlDatasetReader
from allennlp.common.util import ensure_list
from allennlp.common.testing import AllenNlpTestCase


class TestTemplateText2SqlDatasetReader(AllenNlpTestCase):
    def test_reader(self):
        reader = TemplateText2SqlDatasetReader()

        instances = reader.read(str(AllenNlpTestCase.FIXTURES_ROOT / "data" / "text2sql" / "*.json"))
        instances = ensure_list(instances)

        fields = instances[0].fields
        tokens = [t.text for t in fields["tokens"].tokens]
        tags = fields["slot_tags"].labels

        assert tokens == ['how', 'many', 'buttercup', 'kitchen', 'are', 'there', 'in', 'san', 'francisco', '?']
        assert tags == ['O', 'O', 'name0', 'name0', 'O', 'O', 'O', 'city_name0', 'city_name0', 'O']
        assert fields["template"].label == "SELECT COUNT ( * ) FROM LOCATION AS LOCATIONalias0 , RESTAURANT " \
                                           "AS RESTAURANTalias0 WHERE LOCATIONalias0 . CITY_NAME = 'city_name0' " \
                                           "AND RESTAURANTalias0 . ID = LOCATIONalias0 . RESTAURANT_ID AND " \
                                           "RESTAURANTalias0 . NAME = 'name0' ;"

        fields = instances[1].fields
        tokens = [t.text for t in fields["tokens"].tokens]
        tags = fields["slot_tags"].labels
        assert tokens == ['how', 'many', 'chinese', 'restaurants', 'are', 'there', 'in', 'the', 'bay', 'area', '?']
        assert tags == ['O', 'O', 'food_type0', 'O', 'O', 'O', 'O', 'O', 'region0', 'region0', 'O']
        assert fields["template"].label == "SELECT COUNT ( * ) FROM GEOGRAPHIC AS GEOGRAPHICalias0 , RESTAURANT AS "\
                                           "RESTAURANTalias0 WHERE GEOGRAPHICalias0 . REGION = 'region0' AND "\
                                           "RESTAURANTalias0 . CITY_NAME = GEOGRAPHICalias0 . CITY_NAME AND "\
                                           "RESTAURANTalias0 . FOOD_TYPE = 'food_type0' ;"

        fields = instances[2].fields
        tokens = [t.text for t in fields["tokens"].tokens]
        tags = fields["slot_tags"].labels
        assert tokens == ['how', 'many', 'places', 'for', 'chinese', 'food', 'are', 'there', 'in', 'the', 'bay', 'area', '?']
        assert tags == ['O', 'O', 'O', 'O', 'food_type0', 'O', 'O', 'O', 'O', 'O', 'region0', 'region0', 'O']
        assert fields["template"].label == "SELECT COUNT ( * ) FROM GEOGRAPHIC AS GEOGRAPHICalias0 , RESTAURANT AS "\
                                           "RESTAURANTalias0 WHERE GEOGRAPHICalias0 . REGION = 'region0' AND "\
                                           "RESTAURANTalias0 . CITY_NAME = GEOGRAPHICalias0 . CITY_NAME AND "\
                                           "RESTAURANTalias0 . FOOD_TYPE = 'food_type0' ;"

        fields = instances[3].fields
        tokens = [t.text for t in fields["tokens"].tokens]
        tags = fields["slot_tags"].labels
        assert tokens == ['how', 'many', 'chinese', 'places', 'are', 'there', 'in', 'the', 'bay', 'area', '?']
        assert tags == ['O', 'O', 'food_type0', 'O', 'O', 'O', 'O', 'O', 'region0', 'region0', 'O']
        assert fields["template"].label == "SELECT COUNT ( * ) FROM GEOGRAPHIC AS GEOGRAPHICalias0 , RESTAURANT AS "\
                                           "RESTAURANTalias0 WHERE GEOGRAPHICalias0 . REGION = 'region0' AND "\
                                           "RESTAURANTalias0 . CITY_NAME = GEOGRAPHICalias0 . CITY_NAME AND "\
                                           "RESTAURANTalias0 . FOOD_TYPE = 'food_type0' ;"
        fields = instances[4].fields
        tokens = [t.text for t in fields["tokens"].tokens]
        tags = fields["slot_tags"].labels
        assert tokens == ['how', 'many', 'places', 'for', 'chinese', 'are', 'there', 'in', 'the', 'bay', 'area', '?']
        assert tags == ['O', 'O', 'O', 'O', 'food_type0', 'O', 'O', 'O', 'O', 'region0', 'region0', 'O']
        assert fields["template"].label == "SELECT COUNT ( * ) FROM GEOGRAPHIC AS GEOGRAPHICalias0 , RESTAURANT AS "\
                                           "RESTAURANTalias0 WHERE GEOGRAPHICalias0 . REGION = 'region0' AND "\
                                           "RESTAURANTalias0 . CITY_NAME = GEOGRAPHICalias0 . CITY_NAME AND "\
                                           "RESTAURANTalias0 . FOOD_TYPE = 'food_type0' ;"



from typing import List

import torch

from allennlp.nn import util


class RnnStatelet:
    """
    This class keeps track of all of decoder-RNN-related variables that you need during decoding.
    This includes things like the current decoder hidden state, the memory cell (for LSTM
    decoders), the encoder output that you need for computing attentions, and so on.

    This is intended to be used `inside` a ``State``, which likely has other things it has to keep
    track of for doing constrained decoding.

    Parameters
    ----------
    hidden_state : ``torch.Tensor``
        This holds the LSTM hidden state, with shape ``(decoder_output_dim,)`` if the decoder
        has 1 layer and ``(num_layers, decoder_output_dim)`` otherwise.
    memory_cell : ``torch.Tensor``
        This holds the LSTM memory cell, with shape ``(decoder_output_dim,)`` if the decoder has
        1 layer and ``(num_layers, decoder_output_dim)`` otherwise.
    previous_action_embedding : ``torch.Tensor``
        This holds the embedding for the action we took at the last timestep (which gets input to
        the decoder).  Has shape ``(action_embedding_dim,)``.
    attended_input : ``torch.Tensor``
        This holds the attention-weighted sum over the input representations that we computed in
        the previous timestep.  We keep this as part of the state because we use the previous
        attention as part of our decoder cell update.  Has shape ``(encoder_output_dim,)``.
    encoder_outputs : ``List[torch.Tensor]``
        A list of variables, each of shape ``(input_sequence_length, encoder_output_dim)``,
        containing the encoder outputs at each timestep.  The list is over batch elements, and we
        do the input this way so we can easily do a ``torch.cat`` on a list of indices into this
        batched list.

        Note that all of the above parameters are single tensors, while the encoder outputs and
        mask are lists of length ``batch_size``.  We always pass around the encoder outputs and
        mask unmodified, regardless of what's in the grouping for this state.  We'll use the
        ``batch_indices`` for the group to pull pieces out of these lists when we're ready to
        actually do some computation.
    encoder_output_mask : ``List[torch.Tensor]``
        A list of variables, each of shape ``(input_sequence_length,)``, containing a mask over
        question tokens for each batch instance.  This is a list over batch elements, for the same
        reasons as above.
    """
    def __init__(self,
                 hidden_state: torch.Tensor,
                 memory_cell: torch.Tensor,
                 previous_action_embedding: torch.Tensor,
                 attended_input: torch.Tensor,
                 encoder_outputs: List[torch.Tensor],
                 encoder_output_mask: List[torch.Tensor]) -> None:
        self.hidden_state = hidden_state
        self.memory_cell = memory_cell
        self.previous_action_embedding = previous_action_embedding
        self.attended_input = attended_input
        self.encoder_outputs = encoder_outputs
        self.encoder_output_mask = encoder_output_mask

    def __eq__(self, other):
        if isinstance(self, other.__class__):
            return all([
                    util.tensors_equal(self.hidden_state, other.hidden_state, tolerance=1e-5),
                    util.tensors_equal(self.memory_cell, other.memory_cell, tolerance=1e-5),
                    util.tensors_equal(self.previous_action_embedding, other.previous_action_embedding,
                                       tolerance=1e-5),
                    util.tensors_equal(self.attended_input, other.attended_input, tolerance=1e-5),
                    util.tensors_equal(self.encoder_outputs, other.encoder_outputs, tolerance=1e-5),
                    util.tensors_equal(self.encoder_output_mask, other.encoder_output_mask, tolerance=1e-5),
                    ])
        return NotImplemented



from overrides import overrides
import torch
from torch.nn.parameter import Parameter

from allennlp.modules.attention.attention import Attention
from allennlp.nn import Activation


@Attention.register("bilinear")
class BilinearAttention(Attention):
    """
    Computes attention between a vector and a matrix using a bilinear attention function.  This
    function has a matrix of weights ``W`` and a bias ``b``, and the similarity between the vector
    ``x`` and the matrix ``y`` is computed as ``x^T W y + b``.

    Parameters
    ----------
    vector_dim : ``int``
        The dimension of the vector, ``x``, described above.  This is ``x.size()[-1]`` - the length
        of the vector that will go into the similarity computation.  We need this so we can build
        the weight matrix correctly.
    matrix_dim : ``int``
        The dimension of the matrix, ``y``, described above.  This is ``y.size()[-1]`` - the length
        of the vector that will go into the similarity computation.  We need this so we can build
        the weight matrix correctly.
    activation : ``Activation``, optional (default=linear (i.e. no activation))
        An activation function applied after the ``x^T W y + b`` calculation.  Default is no
        activation.
    normalize : ``bool``, optional (default: ``True``)
        If true, we normalize the computed similarities with a softmax, to return a probability
        distribution for your attention.  If false, this is just computing a similarity score.
    """
    def __init__(self,
                 vector_dim: int,
                 matrix_dim: int,
                 activation: Activation = None,
                 normalize: bool = True) -> None:
        super().__init__(normalize)
        self._weight_matrix = Parameter(torch.Tensor(vector_dim, matrix_dim))
        self._bias = Parameter(torch.Tensor(1))
        self._activation = activation or Activation.by_name('linear')()
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self._weight_matrix)
        self._bias.data.fill_(0)

    @overrides
    def _forward_internal(self, vector: torch.Tensor, matrix: torch.Tensor) -> torch.Tensor:
        intermediate = vector.mm(self._weight_matrix).unsqueeze(1)
        return self._activation(intermediate.bmm(matrix.transpose(1, 2)).squeeze(1) + self._bias)
